import unittest
import sys
import os
import socket
from _io import StringIO
# skipcq: BAN-B404
import subprocess

from aminer.util.SecureOSFunctions import secure_open_file, send_annotated_file_descriptor, receive_annoted_file_descriptor
from aminer.util import SecureOSFunctions
from aminer.input.LogStream import UnixSocketLogDataResource
from unit.TestBase import TestBase


class SecureOSFunctionsTestLocal(TestBase):
    """This test class must be run locally due to import problems."""

    opening_socket = 'Opening socket...'
    listening = 'Listening...'

    """This test case is commented out, because it is still not implemented."""
    # '''
    # A file is tried to be opened by using the secure function.
    # '''
    # def test1_secure_open_file(self):
    #   error = sys.stderr = StringIO()
    #   # if an exception is thrown, the test fails.
    #   secure_open_file(b'/etc/aminer/conf-enabled/Readme.txt', os.O_RDONLY)
    #   SecureOSFunctions.no_secure_open_warn_once_flag = True
    #   self.assertEqual(error.getvalue(), '')

    def test2_secure_open_file_relative_path(self):
        """A file is tried to be opened by using a relative path."""
        self.assertRaises(Exception, secure_open_file, 'JsonUtilTest.py', os.O_RDONLY)

    def test3_secure_open_directory(self):
        """A directory is tried to be opened without using the O_Directory flag."""
        error = sys.stderr = StringIO()
        directory = b'/etc/'
        self.assertRaises(Exception, secure_open_file, directory, os.O_RDONLY)
        secure_open_file(directory, os.O_DIRECTORY)
        SecureOSFunctions.no_secure_open_warn_once_flag = True
        self.assertTrue(
            error.getvalue() in ['WARNING: SECURITY: No secure open yet due to missing openat in python!\n', ''])

    def test4sendAnnotatedFileDescriptor(self):
        """A valid annotated file descriptor is to be sent by a socket."""
        sock_name = '/tmp/test4unixSocket.sock'  # skipcq: BAN-B108
        data = b'readmeStream' + b'\x00' + b'You should read these README instructions for better understanding.'

        # skipcq: BAN-B607, BAN-B603
        proc = subprocess.Popen(['python3', 'unit/util/clientTest4.py'])

        if os.path.exists(sock_name):
            os.remove(sock_name)

        # print(self.opening_socket)
        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(sock_name)
        server.listen(1)
        connection = server.accept()[0]
        unix_socket_log_data_resource = UnixSocketLogDataResource(b'unix:///tmp/test4unixSocket.sock', connection.fileno())

        # print(self.listening)
        unix_socket_log_data_resource.fill_buffer()
        self.assertEqual(unix_socket_log_data_resource.buffer, data)
        # print('Data received: %s' % unix_socket_log_data_resource.buffer)

        unix_socket_log_data_resource.update_position(len(unix_socket_log_data_resource.buffer))
        self.assertEqual(unix_socket_log_data_resource.total_consumed_length, 80)
        self.assertEqual(unix_socket_log_data_resource.buffer, b'')

        # print("Shutting down...")
        unix_socket_log_data_resource.close()
        proc.terminate()
        proc.wait()
        # print("Done")

    def test5send_annotated_file_descriptor_invalid_parameters(self):
        """An invalid access is to be performed by using a closed socket."""
        # socket is closed
        fd = secure_open_file(b'/etc/aminer/conf-enabled/Readme.txt', os.O_RDONLY)
        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.assertRaises(OSError, send_annotated_file_descriptor, client, fd, b'readmeStream',
                          b'You should read these README instructions for better understanding.')

    def test6send_logstream_descriptor(self):
        """A valid logstream descriptor is to be sent."""
        sock_name = '/tmp/test6unixSocket.sock'  # skipcq: BAN-B108
        data = b'logstream' + b'\x00' + b'/var/log/syslog'

        # skipcq: BAN-B607, BAN-B603
        subprocess.Popen(['python3', 'unit/util/clientTest6.py'])

        if os.path.exists(sock_name):
            os.remove(sock_name)

        # print(self.opening_socket)
        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(sock_name)
        server.listen(1)
        connection = server.accept()[0]
        unix_socket_log_data_resource = UnixSocketLogDataResource(b'unix:///tmp/test6unixSocket.sock', connection.fileno())

        # print(self.listening)
        unix_socket_log_data_resource.fill_buffer()
        self.assertEqual(unix_socket_log_data_resource.buffer, data)
        # print('Data received: %s' % unix_socket_log_data_resource.buffer)

        unix_socket_log_data_resource.update_position(len(unix_socket_log_data_resource.buffer))
        self.assertEqual(unix_socket_log_data_resource.total_consumed_length, 25)
        self.assertEqual(unix_socket_log_data_resource.buffer, b'')

        # print("Shutting down...")
        unix_socket_log_data_resource.close()
        # print("Done")

    def test7receive_annotated_file_descriptor(self):
        """A valid annotated file descriptor is to be received by a socket."""
        sock_name = '/tmp/test6unixSocket.sock'  # skipcq: BAN-B108
        type_info = b'logstream'
        path = b'/var/log/syslog'
        data = (type_info, path)

        # skipcq: BAN-B607, BAN-B603
        subprocess.Popen(['python3', 'unit/util/clientTest6.py'])

        if os.path.exists(sock_name):
            os.remove(sock_name)

        # print(self.opening_socket)
        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(sock_name)
        server.listen(1)
        connection = server.accept()[0]
        data_tuple = receive_annoted_file_descriptor(connection)

        # print(self.listening)
        self.assertEqual(data_tuple[1], data[0])
        self.assertEqual(data_tuple[2], data[1])
        # print('Data received: (%i, %s, %s)' % data_tuple)
        self.assertEqual(len(data_tuple[1]) + len(data_tuple[2]), 24)
        # print("Done")


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.analysis.PCADetector import PCADetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomalies = []

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomalies.append(evdat)


class PCADetectorTest(TestBase):
    """Unittests for the PCADetector."""

    def test1_normal_pca_detection(self):
        """This test case checks the normal detection of value frequencies using PCA."""
        description = "Test1PCADetector"

        # Initialize detector for analyzing values in one path in time windows of 10 seconds
        test_handler = TestHandler()
        pca_detector = PCADetector(self.aminer_config, ['/value'], [test_handler], 10, 2, 0.9, 3, 'Default', True, output_logline=False)
        self.analysis_context.register_component(pca_detector, description)

        # Prepare log atoms that represent different amounts of values a, b over time
        # Five time windows are used. The first three time windows are used for initializing
        # the count matrix. The fourth window is used to verify the anomaly score computation.
        # The fifth time window is only used to mark the end of the fourth time window.
        # The following log atoms are created:
        #  window 1:
        #   value a: 2 times
        #   value b: 1 time
        #  window 2:
        #   value a: 1 times
        #   value b: 1 time
        #  window 3:
        #   value a: 1 time
        #   value b: 0 times
        #  window 4:
        #   value a: 4 time
        #   value b: 1 time
        #  window 5:
        #   value a: 1 time
        # Start of window 1:
        m_1 = MatchElement('/value', b'a', b'a', None)
        parser_match_1 = ParserMatch(m_1)
        log_atom_1 = LogAtom(b'a', parser_match_1, 1, None)

        m_2 = MatchElement('/value', b'b', b'b', None)
        parser_match_2 = ParserMatch(m_2)
        log_atom_2 = LogAtom(b'b', parser_match_2, 3, None)

        m_3 = MatchElement('/value', b'a', b'a', None)
        parser_match_3 = ParserMatch(m_3)
        log_atom_3 = LogAtom(b'a', parser_match_3, 7, None)

        # Start of window 2:
        m_4 = MatchElement('/value', b'a', b'a', None)
        parser_match_4 = ParserMatch(m_4)
        log_atom_4 = LogAtom(b'a', parser_match_4, 13, None)

        m_5 = MatchElement('/value', b'b', b'b', None)
        parser_match_5 = ParserMatch(m_5)
        log_atom_5 = LogAtom(b'b', parser_match_5, 15, None)

        # Start of window 3:
        m_6 = MatchElement('/value', b'a', b'a', None)
        parser_match_6 = ParserMatch(m_6)
        log_atom_6 = LogAtom(b'b', parser_match_6, 27, None)

        # Start of window 4:
        m_7 = MatchElement('/value', b'a', b'a', None)
        parser_match_7 = ParserMatch(m_7)
        log_atom_7 = LogAtom(b'a', parser_match_7, 33, None)

        m_8 = MatchElement('/value', b'a', b'a', None)
        parser_match_8 = ParserMatch(m_8)
        log_atom_8 = LogAtom(b'a', parser_match_8, 34, None)

        m_9 = MatchElement('/value', b'a', b'a', None)
        parser_match_9 = ParserMatch(m_9)
        log_atom_9 = LogAtom(b'a', parser_match_9, 36, None)

        m_10 = MatchElement('/value', b'a', b'a', None)
        parser_match_10 = ParserMatch(m_10)
        log_atom_10 = LogAtom(b'a', parser_match_10, 37, None)

        m_11 = MatchElement('/value', b'b', b'b', None)
        parser_match_11 = ParserMatch(m_11)
        log_atom_11 = LogAtom(b'b', parser_match_11, 38, None)

        # Start of window 5:
        m_12 = MatchElement('/value', b'a', b'a', None)
        parser_match_12 = ParserMatch(m_12)
        log_atom_12 = LogAtom(b'a', parser_match_12, 45, None)

        # Forward log atoms to detector
        # Log atoms of windows 1 to 3 build up the count matrix
        # Input: log atoms of windows 1 to 3
        # Expected output: No anomalies reported
        pca_detector.receive_atom(log_atom_1)
        pca_detector.receive_atom(log_atom_2)
        pca_detector.receive_atom(log_atom_3)
        pca_detector.receive_atom(log_atom_4)
        pca_detector.receive_atom(log_atom_5)
        pca_detector.receive_atom(log_atom_6)
        self.assertFalse(test_handler.anomalies)

        # Log atoms of window 4 build the count vector for that window
        # Input: b; log atoms of window 4
        # Expected output: No anomalies reported
        pca_detector.receive_atom(log_atom_7)
        pca_detector.receive_atom(log_atom_8)
        pca_detector.receive_atom(log_atom_9)
        pca_detector.receive_atom(log_atom_10)
        pca_detector.receive_atom(log_atom_11)
        self.assertFalse(test_handler.anomalies)
        # At this point, the event count matrix contains the counts from the first three windows
        self.assertEqual(pca_detector.event_count_matrix,
                         [{'/value': {'a': 2, 'b': 1}}, {'/value': {'a': 1, 'b': 1}}, {'/value': {'a': 1, 'b': 0}}])
        # The count vector contains the counts of the fourth window
        self.assertEqual(pca_detector.event_count_vector, {'/value': {'a': 4, 'b': 1}})

        # Log atom of window 5 triggers comparison of count vector from window 4 with PCA
        # Input: log atoms of window 5
        # Expected output: Anomaly reported on count vector of fourth window
        pca_detector.receive_atom(log_atom_12)
        self.assertEqual(test_handler.anomalies, [{'AnalysisComponent': {'AffectedLogAtomPaths': ['/value'],
                         'AffectedLogAtomValues': [['a', 'b']], 'AffectedValueCounts': [[4, 1]], 'AnomalyScore': 9.0}}])
        # Event count matrix is shifted by 1 so that window 0 is removed and window 4 is appended
        self.assertEqual(pca_detector.event_count_matrix,
                         [{'/value': {'a': 1, 'b': 1}}, {'/value': {'a': 1, 'b': 0}}, {'/value': {'a': 4, 'b': 1}}])


if __name__ == "__main__":
    unittest.main()



"""This module defines a generated parser model."""

from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement


def get_model():
    """Return a model to parse Apache Error logs from the AIT-LDS."""
    model = SequenceModelElement("model", [
        FixedDataModelElement("sp1", b"["),
        FixedWordlistDataModelElement("day", [b"Mon", b"Tue", b"Wed", b"Thu", b"Fri", b"Sat", b"Sun"]),
        FixedDataModelElement("sp2", b" "),
        DateTimeModelElement("time", b"%b %d %H:%M:%S.%f %Y"),
        FixedDataModelElement("error_str", b"] [:error] [pid "),
        DecimalIntegerValueModelElement("pid"),
        FixedDataModelElement("sp3", b"] [client "),
        IpAddressDataModelElement("client_ip"),
        FixedDataModelElement("colon", b":"),
        DecimalIntegerValueModelElement("client_port"),
        FixedDataModelElement("php", b"] PHP "),
        FirstMatchModelElement("fphp", [
            SequenceModelElement("warning", [
                FixedDataModelElement("warning_str", b"Warning:  "),
                FirstMatchModelElement("warning", [
                    SequenceModelElement("declaration", [
                        FixedDataModelElement("declaration_str", b"Declaration of "),
                        DelimitedDataModelElement("function", b")"),
                        FixedDataModelElement("compatible_str", b") should be compatible with "),
                        DelimitedDataModelElement("function2", b")"),
                        FixedDataModelElement("compatible_str", b") in "),
                        DelimitedDataModelElement("path", b" "),
                        FixedDataModelElement("compatible_str", b" on line "),
                        DecimalIntegerValueModelElement("line"),
                        FixedDataModelElement("referer_str", b", referer: "),
                        AnyByteDataModelElement("referer")]),
                    SequenceModelElement("system", [
                        FixedDataModelElement("system_str", b"system(): Cannot execute a blank command in "),
                        DelimitedDataModelElement("path", b" "),
                        FixedDataModelElement("compatible_str", b" on line "),
                        DecimalIntegerValueModelElement("line")])])]),
            SequenceModelElement("notice", [
                FixedDataModelElement("notice_str", b"Notice:  Undefined index: "),
                DelimitedDataModelElement("command", b" "),
                FixedDataModelElement("sp", b" in "),
                DelimitedDataModelElement("path", b" "),
                FixedDataModelElement("compatible_str", b" on line "),
                DecimalIntegerValueModelElement("line")]),
            SequenceModelElement("deprecated", [
                FixedDataModelElement("deprecated_str", b"Deprecated:  Methods with the same name as their class "
                                      b"will not be constructors in a future version of PHP; "),
                DelimitedDataModelElement("class", b" "),
                FixedDataModelElement("constructor_str", b" has a deprecated constructor in "),
                DelimitedDataModelElement("path", b" "),
                FixedDataModelElement("compatible_str", b" on line "),
                DecimalIntegerValueModelElement("line"),
                FixedDataModelElement("referer_str", b", referer: "),
                AnyByteDataModelElement("referer"),
                ])])])

    return model



"""This module contains methods which can be executed from the aminerRemoteControl class.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
import aminer
import resource
import os
import shutil
from time import time
from datetime import datetime
import logging
import re
from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.util import PersistenceUtil
from aminer import AnalysisChild, AminerConfig
from aminer.AminerConfig import KEY_PERSISTENCE_PERIOD, KEY_LOG_STAT_LEVEL, KEY_LOG_DEBUG_LEVEL, KEY_LOG_STAT_PERIOD,\
    KEY_RESOURCES_MAX_MEMORY_USAGE, KEY_LOG_PREFIX, KEY_PERSISTENCE_DIR, DEFAULT_PERSISTENCE_DIR, KEY_LOG_SOURCES_LIST, DEBUG_LOG_NAME

attr_str = '"%s": %s,\n'
component_not_found = 'Event history component not found.'


class AminerRemoteControlExecutionMethods:
    """This class defines all possible methods for the remote control."""

    REMOTE_CONTROL_RESPONSE = ''
    ERROR_MESSAGE_RESOURCE_NOT_FOUND = '"Resource \\"%s\\" could not be found."'

    CONFIG_KEY_MAIL_TARGET_ADDRESS = 'MailAlerting.TargetAddress'
    CONFIG_KEY_MAIL_FROM_ADDRESS = 'MailAlerting.FromAddress'
    CONFIG_KEY_MAIL_SUBJECT_PREFIX = 'MailAlerting.SubjectPrefix'
    CONFIG_KEY_MAIL_ALERT_GRACE_TIME = 'MailAlerting.AlertGraceTime'
    CONFIG_KEY_EVENT_COLLECT_TIME = 'MailAlerting.EventCollectTime'
    CONFIG_KEY_ALERT_MIN_GAP = 'MailAlerting.MinAlertGap'
    CONFIG_KEY_ALERT_MAX_GAP = 'MailAlerting.MaxAlertGap'
    CONFIG_KEY_ALERT_MAX_EVENTS_PER_MESSAGE = 'MailAlerting.MaxEventsPerMessage'

    MAIL_CONFIG_PROPERTIES = [CONFIG_KEY_MAIL_TARGET_ADDRESS, CONFIG_KEY_MAIL_FROM_ADDRESS]
    INTEGER_CONFIG_PROPERTY_LIST = [
        CONFIG_KEY_MAIL_ALERT_GRACE_TIME, CONFIG_KEY_EVENT_COLLECT_TIME, CONFIG_KEY_ALERT_MIN_GAP, CONFIG_KEY_ALERT_MAX_GAP,
        CONFIG_KEY_ALERT_MAX_EVENTS_PER_MESSAGE, KEY_PERSISTENCE_PERIOD, KEY_LOG_STAT_LEVEL, KEY_LOG_DEBUG_LEVEL, KEY_LOG_STAT_PERIOD,
        KEY_RESOURCES_MAX_MEMORY_USAGE
    ]
    STRING_CONFIG_PROPERTY_LIST = [
        CONFIG_KEY_MAIL_TARGET_ADDRESS, CONFIG_KEY_MAIL_FROM_ADDRESS, CONFIG_KEY_MAIL_SUBJECT_PREFIX, KEY_LOG_PREFIX
    ]

    def print_response(self, value):
        """Add a value to the response string."""
        self.REMOTE_CONTROL_RESPONSE += str(value)

    def change_config_property(self, analysis_context, property_name, value):
        """Change a config_property in an running aminer instance."""
        result = 0
        config_keys_mail_alerting = [
            self.CONFIG_KEY_MAIL_TARGET_ADDRESS, self.CONFIG_KEY_MAIL_FROM_ADDRESS, self.CONFIG_KEY_MAIL_SUBJECT_PREFIX,
            self.CONFIG_KEY_EVENT_COLLECT_TIME, self.CONFIG_KEY_ALERT_MIN_GAP, self.CONFIG_KEY_ALERT_MAX_GAP,
            self.CONFIG_KEY_ALERT_MAX_EVENTS_PER_MESSAGE, self.CONFIG_KEY_MAIL_ALERT_GRACE_TIME]
        if not isinstance(analysis_context, AnalysisChild.AnalysisContext):
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the analysis_context must be of type %s." % AnalysisChild.AnalysisContext.__class__
            return

        if property_name not in self.INTEGER_CONFIG_PROPERTY_LIST + self.STRING_CONFIG_PROPERTY_LIST:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the property '%s' does not exist in the current config!" % property_name
            return

        if property_name in self.INTEGER_CONFIG_PROPERTY_LIST:
            t = int
        else:
            t = str
        if not isinstance(value, t):
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the value of the property '%s' must be of type %s!" % (property_name, t)
            return

        if property_name in [KEY_PERSISTENCE_DIR, KEY_LOG_SOURCES_LIST]:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the property '%s' can only be changed at " \
                                            "startup in the aminer root process!" % property_name
            return
        if property_name == KEY_RESOURCES_MAX_MEMORY_USAGE:
            result = self.change_config_property_max_memory(analysis_context, value)
        elif property_name in config_keys_mail_alerting:
            result = self.change_config_property_mail_alerting(analysis_context, property_name, value)
        elif property_name in (KEY_LOG_PREFIX, KEY_PERSISTENCE_PERIOD, KEY_LOG_STAT_PERIOD):
            analysis_context.aminer_config.config_properties[property_name] = value
            result = 0
        elif property_name == KEY_LOG_STAT_LEVEL:
            result = self.change_config_property_log_stat_level(analysis_context, value)
        elif property_name == KEY_LOG_DEBUG_LEVEL:
            result = self.change_config_property_log_debug_level(analysis_context, value)
        else:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: property %s could not be changed. Please check the property_name " \
                                            "again." % property_name
            return
        if result == 0:
            msg = "'%s' changed to '%s' successfully." % (property_name, value)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def change_config_property_mail_alerting(self, analysis_context, property_name, value):
        """Change any mail property."""
        is_email = re.compile(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)|^[a-zA-Z0-9]+@localhost$")
        if property_name in self.MAIL_CONFIG_PROPERTIES and not is_email.match(value):
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: MailAlerting.TargetAddress and MailAlerting.FromAddress must be email addresses!"
            return 1
        analysis_context.aminer_config.config_properties[property_name] = value
        for analysis_component_id in analysis_context.get_registered_component_ids():
            component = analysis_context.get_component_by_id(analysis_component_id)
            if component.__class__.__name__ == "DefaultMailNotificationEventHandler":
                setattr(component, property_name, value)
        return 0

    def change_config_property_max_memory(self, analysis_context, max_memory_mb):
        """Change the maximal allowed RAM usage of the aminer instance."""
        try:
            max_memory_mb = int(max_memory_mb)
            if max_memory_mb < 32 and max_memory_mb != -1:
                self.REMOTE_CONTROL_RESPONSE += "FAILURE: it is not safe to run the aminer with less than 32MB RAM."
                return 1
            resource.setrlimit(resource.RLIMIT_AS, (max_memory_mb * 1024 * 1024, resource.RLIM_INFINITY))
            analysis_context.aminer_config.config_properties[KEY_RESOURCES_MAX_MEMORY_USAGE] = max_memory_mb
            return 0
        except ValueError:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: property 'maxMemoryUsage' must be of type Integer!"
            return 1

    def change_config_property_log_stat_level(self, analysis_context, stat_level):
        """Set the statistic logging level."""
        if stat_level in (0, 1, 2):
            analysis_context.aminer_config.config_properties[KEY_LOG_STAT_LEVEL] = stat_level
            AminerConfig.STAT_LEVEL = stat_level
            return 0
        self.REMOTE_CONTROL_RESPONSE += "FAILURE: STAT_LEVEL %d is not allowed. Allowed STAT_LEVEL values are 0, 1, 2." % stat_level
        return 1

    def change_config_property_log_debug_level(self, analysis_context, debug_level):
        """Set the debug log level."""
        if debug_level in (0, 1, 2):
            analysis_context.aminer_config.config_properties[KEY_LOG_DEBUG_LEVEL] = debug_level
            AminerConfig.DEBUG_LEVEL = debug_level
            debug_logger = logging.getLogger(DEBUG_LOG_NAME)
            if debug_level == 0:
                debug_logger.setLevel(logging.ERROR)
            elif debug_level == 1:
                debug_logger.setLevel(logging.INFO)
            else:
                debug_logger.setLevel(logging.DEBUG)
            return 0
        self.REMOTE_CONTROL_RESPONSE += "FAILURE: DEBUG_LEVEL %d is not allowed. Allowed DEBUG_LEVEL values are 0, 1, 2." % debug_level
        return 1

    def change_attribute_of_registered_analysis_component(self, analysis_context, component_name, attribute, value):
        """
        Change a specific attribute of a registered component.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param attribute the name of the attribute to be printed.
        @param value the new value of the attribute.
        """
        attr = getattr(analysis_context.get_component_by_name(component_name), attribute)
        if type(attr) is type(value):
            setattr(analysis_context.get_component_by_name(component_name), attribute, value)
            msg = "'%s.%s' changed from %s to %s successfully." % (component_name, attribute, repr(attr), value)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        else:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: property '%s.%s' must be of type %s!" % (component_name, attribute, type(attr))

    def rename_registered_analysis_component(self, analysis_context, old_component_name, new_component_name):
        """
        Rename an analysis component by removing and readding it to the analysis_context.
        @param analysis_context the analysis context of the aminer.
        @param old_component_name the current name of the component.
        @param new_component_name the new name of the component.
        """
        if type(old_component_name) is not str or type(new_component_name) is not str:
            self.REMOTE_CONTROL_RESPONSE = "FAILURE: the parameters 'old_component_name' and 'new_component_name' must be of type str."
        else:
            component = analysis_context.get_component_by_name(old_component_name)
            if component is None:
                self.REMOTE_CONTROL_RESPONSE += "FAILURE: the component '%s' does not exist." % old_component_name
            else:
                analysis_context.registered_components_by_name[old_component_name] = None
                analysis_context.registered_components_by_name[new_component_name] = component
                msg = "Component '%s' renamed to '%s' successfully." % (old_component_name, new_component_name)
                self.REMOTE_CONTROL_RESPONSE += msg
                logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def print_config_property(self, analysis_context, property_name):
        """
        Print a specific config property.
        @param analysis_context the analysis context of the aminer.
        @param property_name the name of the property to be printed.
        """
        if property_name not in analysis_context.aminer_config.config_properties:
            self.REMOTE_CONTROL_RESPONSE = self.ERROR_MESSAGE_RESOURCE_NOT_FOUND % property_name
            return
        val = analysis_context.aminer_config.config_properties[property_name]
        if isinstance(val, list):
            val = str(val).replace('"False"', 'false').replace('"True"', 'true').replace('"None"', 'null').strip(' ').replace("'", '"')
        else:
            val = str(val).replace('"False"', 'false').replace('"True"', 'true').replace('"None"', 'null').strip(' ')
            if val.isdigit():
                val = int(val)
            elif '.' in val:
                try:
                    val = float(val)
                except ValueError:  # skipcq: FLK-E722
                    pass
        self.REMOTE_CONTROL_RESPONSE = '"%s": %s' % (property_name, val)

    def print_attribute_of_registered_analysis_component(self, analysis_context, component_name, attribute):
        """
        Print a specific attribute of a registered component.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param attribute the name of the attribute to be printed.
        """
        if type(component_name) is not str or type(attribute) is not str:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the parameters 'component_name' and 'attribute' must be of type str."
            return
        if analysis_context.get_component_by_name(component_name) is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the component '%s' does not exist." % component_name
            return
        if hasattr(analysis_context.get_component_by_name(component_name), attribute):
            attr = getattr(analysis_context.get_component_by_name(component_name), attribute, None)
            if isinstance(attr, set):
                attr = list(attr)
            if hasattr(attr, '__dict__') and self.isinstance_aminer_class(attr):
                new_attr = self.get_all_vars(attr, '  ')
                if isinstance(new_attr, str):
                    new_attr = '"%s"' % new_attr
                self.REMOTE_CONTROL_RESPONSE += '"%s.%s": %s' % (component_name, attribute, new_attr)
            elif isinstance(attr, list):
                self.REMOTE_CONTROL_RESPONSE += '"%s.%s": [' % (component_name, attribute)
                for at in attr:
                    if hasattr(at, '__dict__') and self.isinstance_aminer_class(at):
                        new_attr = "\n[\n  " + at.__class__.__name__ + "  {\n" + self.get_all_vars(at, '  ') + "  }\n]"
                    else:
                        if isinstance(at, str):
                            new_attr = '"%s"' % at
                        else:
                            new_attr = str(at)
                    self.REMOTE_CONTROL_RESPONSE += "%s, " % new_attr
                self.REMOTE_CONTROL_RESPONSE = self.REMOTE_CONTROL_RESPONSE.rstrip(", ")
                self.REMOTE_CONTROL_RESPONSE += "]"
            else:
                if attr is None or isinstance(attr, (str, bool)):
                    attr = '"%s"' % attr
                self.REMOTE_CONTROL_RESPONSE += '"%s.%s": %s' % (component_name, attribute, attr)
            self.REMOTE_CONTROL_RESPONSE = self.REMOTE_CONTROL_RESPONSE.replace('"False"', 'false').replace('"True"', 'true').replace(
                '"None"', 'null')
        else:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: the component '%s' does not have an attribute named '%s'." % \
                                            (component_name, attribute)

    def print_current_config(self, analysis_context):
        """
        Print the entire aminer config.
        @param analysis_context the analysis context of the aminer.
        """
        for config_property in analysis_context.aminer_config.config_properties:
            if isinstance(analysis_context.aminer_config.config_properties[config_property], str):
                self.REMOTE_CONTROL_RESPONSE += '"%s": "%s",\n' % (
                    config_property, analysis_context.aminer_config.config_properties[config_property])
            else:
                self.REMOTE_CONTROL_RESPONSE += attr_str % (
                    config_property, analysis_context.aminer_config.config_properties[config_property])
        for component_id in analysis_context.get_registered_component_ids():
            self.REMOTE_CONTROL_RESPONSE += '"%s": {\n' % analysis_context.get_name_by_component(
                analysis_context.get_component_by_id(component_id))
            component = analysis_context.get_component_by_id(component_id)
            self.REMOTE_CONTROL_RESPONSE += self.get_all_vars(component, '  ')
            self.REMOTE_CONTROL_RESPONSE += "},\n\n"
        self.REMOTE_CONTROL_RESPONSE = self.REMOTE_CONTROL_RESPONSE.replace("'", '"').replace('"False"', 'false').replace(
            '"True"', 'true').replace('"None"', 'null').replace('\\"', "'").rstrip(',\n\n\n') + '\n\n'

    def get_all_vars(self, obj, indent):
        """Return all variables in string representation."""
        result = ''
        for var in vars(obj):
            attr = getattr(obj, var, None)
            if attr is not None and isinstance(attr, (tuple, set)):
                attr = list(attr)
            if attr is not None and hasattr(attr, '__dict__') and self.isinstance_aminer_class(attr):
                result += indent + '"%s": {\n' % var + self.get_all_vars(attr, indent + '  ') + indent + "},\n"
            elif isinstance(attr, list):
                for at in attr:
                    if hasattr(at, '__dict__') and self.isinstance_aminer_class(at):
                        result += indent + '"%s": {\n' % var + indent + '  "' + at.__class__.__name__ + \
                                  '": {\n' + self.get_all_vars(at, indent + '    ') + indent + '  ' + "}\n" + indent + '},\n'
                    else:
                        rep = _reformat_attr(attr)
                        result += indent + attr_str % (var, rep)
                        break
            else:
                rep = _reformat_attr(attr)
                result += indent + attr_str % (var, rep)
        return result.rstrip(',\n') + '\n'

    @staticmethod
    def isinstance_aminer_class(obj):
        """Test if an object is of an instance of a aminer class."""
        class_list = [
            aminer.analysis.AtomFilters.SubhandlerFilter, aminer.analysis.AtomFilters.MatchPathFilter,
            aminer.analysis.AtomFilters.MatchValueFilter, aminer.analysis.HistogramAnalysis.LinearNumericBinDefinition,
            aminer.analysis.HistogramAnalysis.BinDefinition, aminer.analysis.HistogramAnalysis.ModuloTimeBinDefinition,
            aminer.analysis.Rules.MatchAction, aminer.analysis.Rules.MatchRule, aminer.analysis.HistogramAnalysis.HistogramData,
            aminer.analysis.TimeCorrelationViolationDetector.CorrelationRule, aminer.analysis.TimeCorrelationDetector.CorrelationFeature,
            aminer.events.EventInterfaces.EventHandlerInterface, aminer.util.History.ObjectHistory]
        for c in class_list:
            if isinstance(obj, c):
                return True
        return False

    def save_current_config(self, analysis_context, destination_file):
        """
        Save the current live config into a file.
        @param analysis_context the analysis context of the aminer.
        @param destination_file the path to the file in which the config is saved.
        """
        if re.match("^(/[^/ ]*)+/?$", destination_file) is not None:
            msg = AminerConfig.save_config(analysis_context, destination_file)
        else:
            msg = f"Exception: {destination_file} is not a valid filename!"
        self.REMOTE_CONTROL_RESPONSE = msg
        logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def persist_all(self):
        """Persist all data by calling the function in PersistenceUtil."""
        PersistenceUtil.persist_all()
        self.REMOTE_CONTROL_RESPONSE = 'OK'
        logging.getLogger(DEBUG_LOG_NAME).info('Called persist_all() via remote control.')

    def create_backup(self, analysis_context):
        """Create a backup with the current datetime string."""
        backup_time = time()
        backup_time_str = datetime.fromtimestamp(backup_time).strftime('%Y-%m-%d-%H-%M-%S')
        persistence_dir = analysis_context.aminer_config.config_properties[KEY_PERSISTENCE_DIR]
        persistence_dir = persistence_dir.rstrip('/')
        backup_path = persistence_dir + '/backup/'
        backup_path_with_date = os.path.join(backup_path, backup_time_str)
        shutil.copytree(persistence_dir, backup_path_with_date, ignore=shutil.ignore_patterns('backup*'))
        msg = 'Created backup %s' % backup_time_str
        self.REMOTE_CONTROL_RESPONSE = 'Created backup %s' % backup_time_str
        logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def list_backups(self, analysis_context):
        """List all available backups from the persistence directory."""
        persistence_dir = analysis_context.aminer_config.config_properties.get(KEY_PERSISTENCE_DIR, DEFAULT_PERSISTENCE_DIR)
        for _dirpath, dirnames, _filenames in os.walk(os.path.join(persistence_dir, 'backup')):
            self.REMOTE_CONTROL_RESPONSE = '"backups": %s' % dirnames
            break
        self.REMOTE_CONTROL_RESPONSE = self.REMOTE_CONTROL_RESPONSE.replace("'", '"')

    def allowlist_event_in_component(self, analysis_context, component_name, event_data, allowlisting_data=None):
        """
        Allowlists one or multiple specific events from the history in the component it occurred in.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param event_data the event_data for the allowlist_event method.
        @param allowlisting_data this data is passed on into the allowlist_event method.
        """
        component = analysis_context.get_component_by_name(component_name)
        if component is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component '%s' does not exist!" % component
            return
        if component.__class__.__name__ not in [
                "EnhancedNewMatchPathValueComboDetector", "MissingMatchPathValueDetector", "NewMatchPathDetector",
                "NewMatchPathValueComboDetector", "NewMatchIdValueComboDetector", "EventCorrelationDetector",
                "NewMatchPathValueDetector"]:
            self.REMOTE_CONTROL_RESPONSE += \
                "FAILURE: component class '%s' does not support allowlisting! Only the following classes support allowlisting: " \
                "EnhancedNewMatchPathValueComboDetector, MissingMatchPathValueDetector, NewMatchPathDetector," \
                " NewMatchIdValueComboDetector, NewMatchPathValueComboDetector, NewMatchPathValueDetector and EventCorrelationDetector." \
                % component.__class__.__name__
            return
        try:
            msg = component.allowlist_event("Analysis.%s" % component.__class__.__name__, event_data, allowlisting_data)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        # skipcq: PYL-W0703
        except Exception as e:
            self.REMOTE_CONTROL_RESPONSE += "Exception: " + repr(e)

    def blocklist_event_in_component(self, analysis_context, component_name, event_data, blocklisting_data=None):
        """
        Blocklists one or multiple specific events from the history in the component it occurred in.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param event_data the event_data for the allowlist_event method.
        @param blocklisting_data this data is passed on into the blocklist_event method.
        """
        component = analysis_context.get_component_by_name(component_name)
        if component is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component '%s' does not exist!" % component
            return
        if component.__class__.__name__ not in ["EventCorrelationDetector"]:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component class '%s' does not support blocklisting! Only the following classes " \
                                            "support blocklisting: EventCorrelationDetector." % component.__class__.__name__
            return
        try:
            msg = component.blocklist_event("Analysis.%s" % component.__class__.__name__, event_data, blocklisting_data)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        # skipcq: PYL-W0703
        except Exception as e:
            self.REMOTE_CONTROL_RESPONSE += "Exception: " + repr(e)

    def print_persistency_event_in_component(self, analysis_context, component_name, event_data):
        """
        Prints the persistency specified in event_data of component_name.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param event_data the event_data for the print_persistency_event method.
        """
        component = analysis_context.get_component_by_name(component_name)
        if component is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component '%s' does not exist!" % component
            return
        if component.__class__.__name__ not in ["EventFrequencyDetector", "MinimalTransitionTimeDetector",
                                                "PathValueTimeIntervalDetector"]:
            self.REMOTE_CONTROL_RESPONSE += \
                "FAILURE: component class '%s' does not support the print_persistency_event! Only the following classes support it: " \
                "EventFrequencyDetector, MinimalTransitionTimeDetector and PathValueTimeIntervalDetector." \
                % component.__class__.__name__
            return
        try:
            msg = component.print_persistence_event("Analysis.%s" % component.__class__.__name__, event_data)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        # skipcq: PYL-W0703
        except Exception as e:
            self.REMOTE_CONTROL_RESPONSE += "Exception: " + repr(e)

    def add_to_persistency_event_in_component(self, analysis_context, component_name, event_data):
        """
        Add information specified in event_data to the persistency of component_name.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param event_data the event_data for the add_to_persistency_event method.
        """
        component = analysis_context.get_component_by_name(component_name)
        if component is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component '%s' does not exist!" % component
            return
        if component.__class__.__name__ not in ["NewMatchPathValueComboDetector", "MinimalTransitionTimeDetector",
                                                "PathValueTimeIntervalDetector"]:
            self.REMOTE_CONTROL_RESPONSE += \
                "FAILURE: component class '%s' does not support the add_to_persistency_event! Only the following classes support it: " \
                "NewMatchPathValueComboDetector, MinimalTransitionTimeDetector and PathValueTimeIntervalDetector." \
                % component.__class__.__name__
            return
        try:
            msg = component.add_to_persistence_event("Analysis.%s" % component.__class__.__name__, event_data)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        # skipcq: PYL-W0703
        except Exception as e:
            self.REMOTE_CONTROL_RESPONSE += "Exception: " + repr(e)

    def remove_from_persistency_event_in_component(self, analysis_context, component_name, event_data):
        """
        Remove information specified in event_data from the persistency of component_name.
        @param analysis_context the analysis context of the aminer.
        @param component_name the name to be registered in the analysis_context.
        @param event_data the event_data for the remove_from_persistency_event method.
        """
        component = analysis_context.get_component_by_name(component_name)
        if component is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component '%s' does not exist!" % component
            return
        if component.__class__.__name__ not in ["MinimalTransitionTimeDetector", "PathValueTimeIntervalDetector"]:
            self.REMOTE_CONTROL_RESPONSE += \
                "FAILURE: component class '%s' does not support the remove_from_persistency_event! Only the following classes support it: "\
                "MinimalTransitionTimeDetector and PathValueTimeIntervalDetector." \
                % component.__class__.__name__
            return
        try:
            msg = component.remove_from_persistence_event("Analysis.%s" % component.__class__.__name__, event_data)
            self.REMOTE_CONTROL_RESPONSE += msg
            logging.getLogger(DEBUG_LOG_NAME).info(msg)
        # skipcq: PYL-W0703
        except Exception as e:
            self.REMOTE_CONTROL_RESPONSE += "Exception: " + repr(e)

    def add_handler_to_atom_filter_and_register_analysis_component(self, analysis_context, atom_handler, component, component_name):
        """
        Add a new component to the analysis_context.
        @param analysis_context the analysis context of the aminer.
        @param atom_handler the registered name of the atom_handler component to add the new component to.
        @param component the component to be added.
        @param component_name the name to be registered in the analysis_context.
        """
        atom_filter = analysis_context.get_component_by_name(atom_handler)
        if atom_filter is None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: atom_handler '%s' does not exist!" % atom_handler
            return
        if analysis_context.get_component_by_name(component_name) is not None:
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: component with same name already registered! (%s)" % component_name
            return
        if not isinstance(component, AtomHandlerInterface):
            self.REMOTE_CONTROL_RESPONSE += "FAILURE: 'component' must implement the AtomHandlerInterface!"
            return
        atom_filter.add_handler(component)
        analysis_context.register_component(component, component_name)
        msg = "Component '%s' added to '%s' successfully." % (component_name, atom_handler)
        self.REMOTE_CONTROL_RESPONSE += msg
        logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def dump_events_from_history(self, analysis_context, history_component_name, dump_event_id):
        """
        Detailed print of a specific event from the history.
        @param analysis_context the analysis context of the aminer.
        @param history_component_name the registered name of the history component.
        @param dump_event_id a numeric id of the events to be printed.
        """
        self.REMOTE_CONTROL_RESPONSE = None
        history_handler = analysis_context.get_component_by_name(history_component_name)
        if history_handler is None:
            self.REMOTE_CONTROL_RESPONSE = component_not_found
        else:
            history_data = history_handler.get_history()
            result_string = 'FAIL: not found'
            for event_pos in enumerate(history_data):
                event_id, event_type, event_message, sorted_log_lines, event_data, _event_source = history_data[event_pos]
                if event_id != dump_event_id:
                    continue
                append_log_lines_flag = True
                result_string = 'OK\nEvent %d: %s (%s)' % (event_id, event_message, event_type)
                if event_type == 'Analysis.NewMatchPathDetector':
                    result_string += '\n  Logline: %s' % (sorted_log_lines[0],)
                elif event_type == 'Analysis.NewMatchPathValueComboDetector':
                    result_string += '\nParser match:\n' + event_data[0].parser_match.matchElement.annotate_match('  ')
                elif event_type == 'Analysis.AllowlistViolationDetector':
                    result_string += '\nParser match:\n' + event_data.parser_match.matchElement.annotate_match('  ')
                elif event_type == 'ParserModel.UnparsedData':
                    result_string += '\n  Unparsed line: %s' % sorted_log_lines[0]
                    append_log_lines_flag = False
                else:
                    result_string += '\n  Data: %s' % str(event_data)

                if append_log_lines_flag and (sorted_log_lines is not None) and (len(sorted_log_lines) != 0):
                    result_string += '\n  Log lines:\n    %s' % '\n    '.join(sorted_log_lines)
                break
            self.REMOTE_CONTROL_RESPONSE = result_string
            logging.getLogger(DEBUG_LOG_NAME).info(result_string)

    def ignore_events_from_history(self, analysis_context, history_component_name, event_ids):
        """
        Ignore one or multiple specific events from the history. These ignores do not affect the components itself.
        @param analysis_context the analysis context of the aminer.
        @param history_component_name the registered name of the history component.
        @param event_ids a list of numeric ids of the events to be ignored.
        """
        history_handler = analysis_context.get_component_by_name(history_component_name)
        if history_handler is None:
            self.REMOTE_CONTROL_RESPONSE = component_not_found
            return
        history_data = history_handler.get_history()
        id_spec_list = []
        for element in event_ids:
            if isinstance(element, list):
                id_spec_list.append(element)
        delete_count = 0
        event_pos = 0
        while event_pos < len(history_data):
            event_id, _event_type, _event_message, _sorted_log_lines, _event_data, _event_source = history_data[event_pos]
            may_delete_flag = False
            if event_id in event_ids:
                may_delete_flag = True
            else:
                for id_range in id_spec_list:
                    if id_range[0] <= event_id <= id_range[1]:
                        may_delete_flag = True
            if may_delete_flag:
                history_data[:] = history_data[:event_pos] + history_data[event_pos + 1:]
                delete_count += 1
            else:
                event_pos += 1
        msg = 'OK\n%d elements ignored' % delete_count
        self.REMOTE_CONTROL_RESPONSE = msg
        logging.getLogger(DEBUG_LOG_NAME).info(msg)

    def list_events_from_history(self, analysis_context, history_component_name, max_event_count=None):
        """
        List the latest events of a specific history component.
        @param analysis_context the analysis context of the aminer.
        @param history_component_name the registered name of the history component.
        @param max_event_count the number of the newest events to be listed.
        """
        history_handler = analysis_context.get_component_by_name(history_component_name)
        if history_handler is None:
            self.REMOTE_CONTROL_RESPONSE = component_not_found
        else:
            history_data = history_handler.get_history()
            max_events = len(history_data)
            if max_event_count is None or max_events < max_event_count:
                max_event_count = max_events
            result_string = 'OK'
            for event_id, _event_type, event_message, sorted_log_lines, _event_data, _event_source in history_data[:max_event_count]:
                result_string += ('\nEvent %d: %s; Log data: %s' % (event_id, event_message, repr(sorted_log_lines)))[:240]
            self.REMOTE_CONTROL_RESPONSE = result_string

    def allowlist_events_from_history(self, analysis_context, history_component_name, id_spec_list, allowlisting_data=None):
        """
        Allowlists one or multiple specific events from the history in the component it occurred in.
        @param analysis_context the analysis context of the aminer.
        @param history_component_name the registered name of the history component.
        @param id_spec_list a list of numeric ids of the events to be allowlisted.
        @param allowlisting_data this data is passed on into the allowlist_event method.
        """
        from aminer.events.EventInterfaces import EventSourceInterface
        history_handler = analysis_context.get_component_by_name(history_component_name)
        if history_handler is None:
            self.REMOTE_CONTROL_RESPONSE = component_not_found
            return
        if id_spec_list is None or not isinstance(id_spec_list, list):
            self.REMOTE_CONTROL_RESPONSE = \
                'Request requires remote_control_data with ID specification list and optional allowlisting information.'
            return
        history_data = history_handler.get_history()
        result_string = ''
        lookup_count = 0
        event_pos = 0
        while event_pos < len(history_data):
            event_id, event_type, _event_message, sorted_log_lines, event_data, event_source = history_data[event_pos]
            found_flag = False
            if event_id in id_spec_list:
                found_flag = True
            else:
                for id_range in id_spec_list:
                    if isinstance(id_range, list) and (id_range[0] <= event_id <= id_range[1]):
                        found_flag = True
            if not found_flag:
                event_pos += 1
                continue
            lookup_count += 1
            allowlisted_flag = False
            if isinstance(event_source, EventSourceInterface):
                # This should be the default for all detectors.
                try:
                    message = event_source.allowlist_event(
                        event_type, sorted_log_lines, event_data, allowlisting_data)
                    result_string += 'OK %d: %s\n' % (event_id, message)
                    logging.getLogger(DEBUG_LOG_NAME).info(result_string)
                    allowlisted_flag = True
                except NotImplementedError:
                    result_string += 'FAIL %d: component does not support allowlisting.' % event_id
                # skipcq: PYL-W0703
                except Exception as wl_exception:
                    result_string += 'FAIL %d: %s\n' % (event_id, str(wl_exception))
            elif event_type == 'Analysis.AllowlistViolationDetector':
                result_string += 'FAIL %d: No automatic modification of allowlist rules, manual changes required\n' % event_id
                allowlisted_flag = True
            elif event_type == 'ParserModel.UnparsedData':
                result_string += 'FAIL %d: No automatic modification of parsers yet\n' % event_id
            else:
                result_string += 'FAIL %d: Unsupported event type %s\n' % (event_id, event_type)
            if allowlisted_flag:
                # Clear the allowlisted event.
                history_data[:] = history_data[:event_pos] + history_data[event_pos + 1:]
            else:
                event_pos += 1
        if lookup_count == 0:
            result_string = 'FAIL: Not a single event ID from specification found'
        self.REMOTE_CONTROL_RESPONSE = result_string

    def reopen_event_handler_streams(self, analysis_context):
        """Reopen all StreamPrinterEventHandler streams for log rotation."""
        analysis_context.close_event_handler_streams(analysis_context.atomizer_factory.event_handler_list, reopen=True)
        msg = "Reopened all StreamPrinterEventHandler streams."
        self.REMOTE_CONTROL_RESPONSE = msg
        logging.getLogger(DEBUG_LOG_NAME).info(msg)


def _repr_recursive(attr):
    """
    Return a valid JSON representation of an config attribute with the types list, dict, set or tuple.
    @param attr the attribute to be represented.
    """
    if attr is None:
        return None
    if isinstance(attr, (bool, type(AminerConfig))):
        rep = str(attr)
    elif isinstance(attr, (int, str, float)):
        rep = attr
    elif isinstance(attr, bytes):
        rep = attr.decode()
    elif isinstance(attr, (list, tuple, set)):
        if isinstance(attr, (tuple, set)):
            attr = list(attr)
        for i, a in enumerate(attr):
            attr[i] = _repr_recursive(a)
        rep = str(attr).replace('\\"', "'").replace("'[", "[").replace("]'", "]").replace("'", '"').replace('"False"', 'false').replace(
            '"True"', 'true').replace('"None"', 'null')
    elif isinstance(attr, dict):
        new_attr = {}
        for key in attr.keys():
            value = _repr_recursive(key)
            if isinstance(value, str):
                value = value.replace('\\"', "'")
            new_attr[str(key)] = value
        rep = str(new_attr).replace("'[", "[").replace("]'", "]")
    else:
        rep = attr.__class__.__name__
    return rep


def _reformat_attr(attr):
    """
    Return a valid JSON representation of an config attribute with any type.
    If the type is list, dict, set or tuple _repr_recursive is called.
    @param attr the attribute to be represented.
    """
    if type(attr) in (int, str, float, bool, type(AminerConfig), type(None)):
        rep = str(attr)
    elif isinstance(attr, bytes):
        rep = attr.decode()
    elif isinstance(attr, (list, dict, set, tuple)):
        rep = _repr_recursive(attr)
    else:
        rep = attr.__class__.__name__

    if rep.startswith("'") and rep.endswith("'") and rep.count("'") == 2:
        rep = rep.replace("'", '"')
    elif rep.strip('"').startswith("'") and rep.strip('"').endswith("'") and rep.strip('"').count("'") == 2:
        rep = rep.strip('"').replace("'", '"')
    else:
        rep = rep.strip('"').replace("'", '\\"')
    if not isinstance(attr, (list, dict, tuple, set)) and not rep.startswith('"') and not rep.isdecimal():
        try:
            float(rep)
        except ValueError:  # skipcq: FLK-E722
            rep = '"%s"' % rep
    return rep



import unittest
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase, DummyMatchContext, DummyFixedDataModelElement


class FirstDataModelElementTest(TestBase):
    """Unittests for the FirstDataModelElement."""

    id_ = "first"
    path = "path"
    me1 = DummyFixedDataModelElement("me1", b"The first fixed string.")
    me2 = DummyFixedDataModelElement("me2", b"Random string23.")
    me3 = DummyFixedDataModelElement("me3", b"Random string2")
    children = [me1, me2, me3]

    def test1get_match_element_valid_match(self):
        """Parse matching substring from MatchContext and check if the MatchContext was updated with all characters."""
        data = b"The first fixed string. Random string23."
        value = b"The first fixed string."
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_+"/me1", self.path, value, value, None)

        data = b"Random string23. Random string23."
        value = b"Random string23."
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_ + "/me2", self.path, value, value, None)

        data = b"Random string2 Random string23."
        value = b"Random string2"
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_ + "/me3", self.path, value, value, None)

        data = b"Random string24. Random string23."
        value = b"Random string2"
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_ + "/me3", self.path, value, value, None)

    def test2get_match_element_no_match(self):
        """Parse not matching substring from MatchContext and check if the MatchContext was not changed."""
        data = b"some none matching string"
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"The first fixed string"
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"Random string42"
        match_context = DummyMatchContext(data)
        first_match_me = FirstMatchModelElement(self.id_, self.children)
        match_element = first_match_me.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test3element_id_input_validation(self):
        """Check if element_id is validated."""
        self.assertRaises(ValueError, FirstMatchModelElement, "", self.children)  # empty element_id
        self.assertRaises(TypeError, FirstMatchModelElement, None, self.children)  # None element_id
        self.assertRaises(TypeError, FirstMatchModelElement, b"path", self.children)  # bytes element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, True, self.children)  # boolean element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, 123, self.children)  # integer element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, 123.22, self.children)  # float element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, {"id": "path"}, self.children)  # dict element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, ["path"], self.children)  # list element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, [], self.children)  # empty list element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, (), self.children)  # empty tuple element_id is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, set(), self.children)  # empty set element_id is not allowed

    def test4children_input_validation(self):
        """Check if children is validated."""
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, "path")  # string children
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, None)  # None children
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, b"path")  # bytes children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, True)  # boolean children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, 123)  # integer children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, 123.22)  # float children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, {"id": "path"})  # dict children is not allowed
        # list children with no ModelElementInterface elements is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, ["path"])
        self.assertRaises(ValueError, FirstMatchModelElement, self.id_, [])  # empty list children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, ())  # empty tuple children is not allowed
        self.assertRaises(TypeError, FirstMatchModelElement, self.id_, set())  # empty set children is not allowed

    def test5get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = FirstMatchModelElement(self.id_, self.children)
        data = b"abcdefghijklmnopqrstuvwxyz.!?"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""This module defines a generated parser model."""

from aminer.parsing.DateTimeModelElement import DateTimeModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement


def get_model():
    """Return a model to parse Apache Access logs from the AIT-LDS."""
    alphabet = b"!'#$%&\"()*+,-./0123456789:;<>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ\\^_`abcdefghijklmnopqrstuvwxyz{|}~=[]"

    model = SequenceModelElement("model", [
        FirstMatchModelElement("client_ip", [
            IpAddressDataModelElement("client_ip"),
            FixedDataModelElement("localhost", b"::1")
            ]),
        FixedDataModelElement("sp1", b" "),
        VariableByteDataModelElement("client_id", alphabet),
        FixedDataModelElement("sp2", b" "),
        VariableByteDataModelElement("user_id", alphabet),
        FixedDataModelElement("sp3", b" ["),
        DateTimeModelElement("time", b"%d/%b/%Y:%H:%M:%S%z"),
        FixedDataModelElement("sp4", b'] "'),
        FirstMatchModelElement("fm", [
            FixedDataModelElement("dash", b"-"),
            SequenceModelElement("request", [
                FixedWordlistDataModelElement("method", [
                    b"GET", b"POST", b"PUT", b"HEAD", b"DELETE", b"CONNECT", b"OPTIONS", b"TRACE", b"PATCH"]),
                FixedDataModelElement("sp5", b" "),
                DelimitedDataModelElement("request", b" ", b"\\"),
                FixedDataModelElement("sp6", b" "),
                DelimitedDataModelElement("version", b'"'),
                ])
            ]),
        FixedDataModelElement("sp7", b'" '),
        DecimalIntegerValueModelElement("status_code"),
        FixedDataModelElement("sp8", b" "),
        DecimalIntegerValueModelElement("content_size"),
        OptionalMatchModelElement(
            "combined", SequenceModelElement("combined", [
                FixedDataModelElement("sp9", b' "'),
                DelimitedDataModelElement("referer", b'"', b"\\"),
                FixedDataModelElement("sp10", b'" "'),
                DelimitedDataModelElement("user_agent", b'"', b"\\"),
                FixedDataModelElement("sp11", b'"'),
                ]))
        ])

    return model



"""
This module defines a model element that takes any string up to a specific delimiter string.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ModelElementInterface import ModelElementInterface
import re


class DelimitedDataModelElement(ModelElementInterface):
    """Find a string delimited by given non-escaped delimiter string, possibly a match of zero byte length."""

    def __init__(self, element_id: str, delimiter: bytes, escape: bytes = None, consume_delimiter: bool = False):
        """
        Initialize the ModelElement.
        @param element_id an identifier for the ModelElement which is shown in the path.
        @param delimiter a non-escaped delimiter string to search for.
        @param escape a character to escape in the string.
        @param consume_delimiter True if the delimiter character should also be consumed.
        """
        super().__init__(element_id, delimiter=delimiter, escape=escape, consume_delimiter=consume_delimiter)

    def get_match_element(self, path: str, match_context):
        """
        Find the maximum number of bytes before encountering the non-escaped delimiter.
        @return a match when at least one byte was found but not the delimiter itself.
        """
        data = match_context.match_data
        match_len = -1
        if self.escape is None:
            search = re.search(re.escape(self.delimiter), data)
            if search is not None:
                match_len = search.start()
        else:
            search = re.search(rb"(?<!" + re.escape(self.escape) + rb")" + re.escape(self.delimiter), data)
            if search is not None:
                match_len = search.start()
        if match_len < 1:
            return None
        match_data = data[:match_len + len(self.delimiter) * (self.consume_delimiter is True)]
        match_context.update(match_data)
        return MatchElement(f"{path}/{self.element_id}", match_data, match_data, None)



"""
This file contains interface definition useful implemented by classes in this directory and for use from code outside this directory.
All classes are defined in separate files, only the namespace references are added here to simplify the code.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.

"""

import abc


class EventHandlerInterface(metaclass=abc.ABCMeta):
    """
    This is the common interface of all components that can be notified on significant log data mining events.
    To avoid interference with the analysis process, the listener may only perform fast actions within the call. Longer running tasks have
    to be performed asynchronously.
    """

    @abc.abstractmethod
    def receive_event(self, event_type, event_message, sorted_loglines, event_data, log_atom, event_source):
        """
        Receive information about a detected event.
        @param event_type is a string with the event type class this event belongs to. This information can be used to interpret
               type-specific event_data objects. Together with the eventMessage and sorted_loglines, this can be used to create generic log
               messages.
        @param event_message the first output line of the event.
        @param sorted_loglines sorted list of log lines that were considered when generating the event, as far as available to the time
               of the event. The list has to contain at least one line.
        @param event_data type-specific event data object, should not be used unless listener really knows about the event_type.
        @param log_atom the log atom which produced the event.
        @param event_source reference to detector generating the event.
        """


class EventSourceInterface(metaclass=abc.ABCMeta):
    """
    This is the common interface of all event sources.
    Component not implementing this interface may still emit events without support for callbacks.
    """

    @abc.abstractmethod
    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws NotImplementedError if this source does not support allowlisting per se
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """

    @staticmethod
    def get_weight_analysis_field_path():
        """Return the path to the list in the output of the detector which is weighted by the ScoringEventHandler."""
        return []

    @staticmethod
    def get_weight_output_field_path():
        """Return the path where the ScoringEventHandler adds the scorings in the output of the detector."""
        return []



import unittest
import sys
import io
from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
import time
from aminer.util import PersistenceUtil
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.analysis.NewMatchPathValueComboDetector import NewMatchPathValueComboDetector
from unit.TestBase import TestBase


class PersistenceUtilTest(TestBase):
    """Unittests for the PersistenceUtil class."""

    string = b'25537 uid=2'

    match_context_fixed_dme = MatchContext(b' pid=')
    fixed_dme = FixedDataModelElement('s1', b' pid=')
    match_element_fixed_dme = fixed_dme.get_match_element("", match_context_fixed_dme)

    match_context_decimal_integer_value_me = MatchContext(string)
    decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                               DecimalIntegerValueModelElement.PAD_TYPE_NONE)
    match_element_decimal_integer_value_me = decimal_integer_value_me.get_match_element("", match_context_decimal_integer_value_me)

    fixed_dme = FixedDataModelElement('s1', string)

    decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                               DecimalIntegerValueModelElement.PAD_TYPE_NONE)

    match_context_first_match_me = MatchContext(string)
    first_match_me = FirstMatchModelElement('f1', [fixed_dme, decimal_integer_value_me])
    match_element_first_match_me = first_match_me.get_match_element('first', match_context_first_match_me)

    match_context_first_match_me2 = MatchContext(string)
    first_match_me2 = FirstMatchModelElement('f2', [decimal_integer_value_me, fixed_dme])
    match_element_first_match_me2 = first_match_me2.get_match_element('second', match_context_first_match_me2)

    def test1persist_multiple_objects_of_single_class(self):
        """In this test case multiple instances of one class are to be persisted and loaded."""
        description = "Test1PersistenceUtil"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True)
        self.analysis_context.register_component(new_match_path_detector, description)

        t = time.time()
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        log_atom_decimal_integer_value_me = LogAtom(self.match_context_decimal_integer_value_me.match_data,
                                                    ParserMatch(self.match_element_decimal_integer_value_me), t, new_match_path_detector)
        new_match_path_detector.receive_atom(log_atom_fixed_dme)
        new_match_path_detector.receive_atom(log_atom_decimal_integer_value_me)

        other_new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'otherDetector', True)
        self.analysis_context.register_component(other_new_match_path_detector, description + "2")
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, other_new_match_path_detector)
        other_new_match_path_detector.receive_atom(log_atom_fixed_dme)

        PersistenceUtil.persist_all()
        persistence_data = PersistenceUtil.load_json(new_match_path_detector.persistence_file_name)
        self.assertTrue(
            persistence_data in ([self.match_element_fixed_dme.get_path(), self.match_element_decimal_integer_value_me.get_path()], [
                self.match_element_decimal_integer_value_me.get_path(), self.match_element_fixed_dme.get_path()]))
        self.assertEqual(PersistenceUtil.load_json(other_new_match_path_detector.persistence_file_name), [
            self.match_element_fixed_dme.get_path()])

    def test2persist_multiple_objects_of_multiple_class(self):
        """In this test case multiple instances of multiple classes are to be persisted and loaded."""
        description = "Test2PersistenceUtil"
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default2', True)
        self.analysis_context.register_component(new_match_path_detector, description)

        t = time.time()
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, new_match_path_detector)
        log_atom_decimal_integer_value_me = LogAtom(self.match_context_decimal_integer_value_me.match_data,
                                                    ParserMatch(self.match_element_decimal_integer_value_me), t, new_match_path_detector)
        new_match_path_detector.receive_atom(log_atom_fixed_dme)
        new_match_path_detector.receive_atom(log_atom_decimal_integer_value_me)

        other_new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'otherDetector2',
                                                             True)
        self.analysis_context.register_component(other_new_match_path_detector, description + "2")
        log_atom_fixed_dme = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_fixed_dme), t, other_new_match_path_detector)
        other_new_match_path_detector.receive_atom(log_atom_fixed_dme)

        new_match_path_value_combo_detector = NewMatchPathValueComboDetector(self.aminer_config, ['first/f1/s1'],
                                                                             [self.stream_printer_event_handler], 'Default', False, True)
        self.analysis_context.register_component(new_match_path_value_combo_detector, description + "3")
        log_atom_sequence_me = LogAtom(self.fixed_dme.fixed_data, ParserMatch(self.match_element_first_match_me), t,
                                       new_match_path_value_combo_detector)
        new_match_path_value_combo_detector.receive_atom(log_atom_sequence_me)

        PersistenceUtil.persist_all()
        persistence_data = PersistenceUtil.load_json(new_match_path_detector.persistence_file_name)
        self.assertTrue(
            persistence_data in ([self.match_element_fixed_dme.get_path(), self.match_element_decimal_integer_value_me.get_path()], [
                self.match_element_decimal_integer_value_me.get_path(), self.match_element_fixed_dme.get_path()]))
        self.assertEqual(PersistenceUtil.load_json(other_new_match_path_detector.persistence_file_name),
                         [self.match_element_fixed_dme.get_path()])
        self.assertEqual(PersistenceUtil.load_json(new_match_path_value_combo_detector.persistence_file_name),
                         ([[log_atom_sequence_me.raw_data]]))

    def test3_no_unique_persistence_id(self):
        """Check if a warning is printed if the same persistence_id is used for the same component type."""
        old_stderr = sys.stderr
        new_stderr = io.StringIO()
        sys.stderr = new_stderr
        PersistenceUtil.SKIP_PERSISTENCE_ID_WARNING = False

        NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True)
        NewMatchPathValueComboDetector(self.aminer_config, ['first/f1/s1'], [self.stream_printer_event_handler], 'Default', False, True)
        self.assertEqual('', new_stderr.getvalue())

        NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', True)
        self.assertEqual('Warning: Detectors of type NewMatchPathDetector use the persistence_id "Default" multiple times. Please assign a'
                         ' unique persistence_id for every component.\n', new_stderr.getvalue())
        new_stderr.seek(0)
        new_stderr.truncate(0)

        NewMatchPathValueComboDetector(self.aminer_config, ['first/f1/s1'], [self.stream_printer_event_handler], 'Default', False, True)
        self.assertEqual('Warning: Detectors of type NewMatchPathValueComboDetector use the persistence_id "Default" multiple times. Please'
                         ' assign a unique persistence_id for every component.\n', new_stderr.getvalue())
        sys.stderr = old_stderr


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.events.Utils import VolatileLogarithmicBackoffEventHistory
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from time import time
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class UtilsTest(TestBase):
    """Unittests for the Utils."""

    pid = b' pid='
    test = 'Test.%s'
    match_s1 = 'match/s1'
    match_s2 = 'match/s2'
    new_val = 'New value for pathes %s, %s: %s '

    def test1add_multiple_objects(self):
        """In this test case multiple events are received by the VolatileLogarithmicBackoffEventHistory."""
        volatile_logarithmic_backoff_event_history = VolatileLogarithmicBackoffEventHistory(10)

        match_context = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element = fixed_dme.get_match_element("match", match_context)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element), t, self)
        message = self.new_val % (self.match_s1, self.match_s2, repr(match_element.match_object))

        volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
            log_atom.raw_data, log_atom.raw_data], None, log_atom, self)
        self.assertEqual(volatile_logarithmic_backoff_event_history.get_history(), [
            (0, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None, log_atom, self)])

        volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
            log_atom.raw_data, log_atom.raw_data], None, log_atom, self)
        self.assertEqual(volatile_logarithmic_backoff_event_history.get_history(), [
            (0, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None, log_atom, self),
            (1, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None, log_atom, self)])

    def test2add_no_objects(self):
        """In this test case no events are received by the VolatileLogarithmicBackoffEventHistory."""
        volatile_logarithmic_backoff_event_history = VolatileLogarithmicBackoffEventHistory(10)
        self.assertEqual(volatile_logarithmic_backoff_event_history.get_history(), [])

    def test3event_data_not_log_atom(self):
        """In this test case the EventHandler receives no logAtom from the test class and the output should not contain the log time."""
        volatile_logarithmic_backoff_event_history = VolatileLogarithmicBackoffEventHistory(10)

        match_context = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element = fixed_dme.get_match_element("match", match_context)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element), t, self)
        message = self.new_val % (self.match_s1, self.match_s2, repr(match_element.match_object))

        volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
            log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)
        self.assertEqual(
            volatile_logarithmic_backoff_event_history.get_history(), [
                (0, self.test % self.__class__.__name__, message, [
                    log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)])

    def test4max_items_overflow(self):
        """In this test case more events than the VolatileLogarithmicBackoffEventHistory can handle are received."""
        deviation = 0.05
        size = 100000
        msg = "%s=%f is not between %f and %f"

        match_context = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element = fixed_dme.get_match_element("match", match_context)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element), t, self)
        message = self.new_val % (self.match_s1, self.match_s2, repr(match_element.match_object))
        first = 0
        second = 0
        third = 0
        fourth = 0

        for _ in range(size):
            volatile_logarithmic_backoff_event_history = VolatileLogarithmicBackoffEventHistory(2)
            volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
                log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)

            volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
                log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)

            volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
                log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)

            volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
                log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)

            volatile_logarithmic_backoff_event_history.receive_event(self.test % self.__class__.__name__, message, [
                log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)

            history = volatile_logarithmic_backoff_event_history.get_history()
            if history == [(0, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None,
                           log_atom.get_parser_match(), self), (4, self.test % self.__class__.__name__, message,
                           [log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)]:
                first += 1
            elif history == [(1, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None,
                             log_atom.get_parser_match(), self), (4, self.test % self.__class__.__name__, message,
                             [log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)]:
                second += 1
            elif history == [(2, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None,
                             log_atom.get_parser_match(), self), (4, self.test % self.__class__.__name__, message,
                             [log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)]:
                third += 1
            elif history == [(3, self.test % self.__class__.__name__, message, [log_atom.raw_data, log_atom.raw_data], None,
                             log_atom.get_parser_match(), self), (4, self.test % self.__class__.__name__, message,
                             [log_atom.raw_data, log_atom.raw_data], None, log_atom.get_parser_match(), self)]:
                fourth += 1
        val = 0.5 * 0.5 * 0.5
        minimum = size * val * (1 - deviation)
        maximum = size * val * (1 + deviation)
        self.assertTrue(minimum <= first <= maximum, msg % ("first", first, minimum, maximum))

        val = 0.5 * 0.5 * 0.5
        minimum = size * val * (1 - deviation)
        maximum = size * val * (1 + deviation)
        self.assertTrue(minimum <= second <= maximum, msg % ("second", second, minimum, maximum))

        val = 2 * 0.5 * 0.5 * 0.5
        minimum = size * val * (1 - deviation)
        maximum = size * val * (1 + deviation)
        self.assertTrue(minimum <= third <= maximum, msg % ("third", third, minimum, maximum))

        val = 0.5
        minimum = size * val * (1 - deviation)
        maximum = size * val * (1 + deviation)
        self.assertTrue(minimum <= fourth <= maximum, msg % ("fourth", fourth, minimum, maximum))


if __name__ == "__main__":
    unittest.main()



"""
This module defines a factory for instanciating line atomizers.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.input.InputInterfaces import AtomizerFactory
from aminer.input.ByteStreamLineAtomizer import ByteStreamLineAtomizer


class SimpleByteStreamLineAtomizerFactory(AtomizerFactory):
    """
    This factory just creates the same atomizer for each new resource.
    All parsed and unparsed atoms are delivered via two lists of handlers.
    """

    def __init__(
            self, parsing_model, atom_handler_list, event_handler_list, default_timestamp_path_list=None, eol_sep=b'\n', json_format=False):
        """
        Create the factory to forward data and events to the given lists for each newly created atomizer.
        @param default_timestamp_path_list if not empty list, the value of this timestamp field is extracted from parsed atoms and stored
        as default timestamp for that atom.
        """
        self.parsing_model = parsing_model
        self.atom_handler_list = atom_handler_list
        self.event_handler_list = event_handler_list
        if default_timestamp_path_list is None:
            self.default_timestamp_path_list = []
        else:
            self.default_timestamp_path_list = default_timestamp_path_list
        self.eol_sep = eol_sep
        self.json_format = json_format

    def get_atomizer_for_resource(self, resource_name):
        """
        Get an atomizer for a given resource.
        @param resource_name the resource name for atomizer selection is ignored in this type of factory.
        @return a StreamAtomizer object
        """
        return ByteStreamLineAtomizer(self.parsing_model, self.atom_handler_list, self.event_handler_list, 1 << 16,
                                      self.default_timestamp_path_list, self.eol_sep, self.json_format)



import unittest
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase, DummyMatchContext


class DelimitedDataModelElementTest(TestBase):
    """Unittests for the DelimitedDataModelElement."""

    id_ = "delimited"
    path = "path"
    delimiter = b","

    def test1get_match_element_single_char(self):
        """A single character is used as delimiter and not consumed (consume_delimiter=False)."""
        data = b"this is a match context.\n"

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"a")
        value = b"this is "
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"c")
        value = b"this is a mat"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"e")
        value = b"this is a match cont"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"\n")
        value = b"this is a match context."
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

    def test2get_match_element_single_char_no_match(self):
        """A single character is used as delimiter and not matched."""
        data = b"this is a match context.\n"
        for char in "bdfgjklpqruvwyz":
            delimited_data_model_element = DelimitedDataModelElement(self.id_, char.encode())
            match_context = DummyMatchContext(data)
            match_element = delimited_data_model_element.get_match_element(self.path, match_context)
            self.compare_no_match_results(data, match_element, match_context)

    def test3delimiter_string(self):
        """In this test case a whole string is searched for in the match_data and  it is not consumed (consume_delimiter=False)."""
        data = b"this is a match context.\n"

        value = b"this"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b" is")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"th"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"is")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"this is a match "
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"context.\n")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"t"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"his is a match context.\n")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

    def test4delimiter_string_no_match(self):
        """In this test case a whole string is searched for in the match_data with no match."""
        data = b"this is a match context.\n"

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"other data")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"isa")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"context\n")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"this is a match context.\n")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test5special_characters_escape(self):
        """In this test case special character escaping is tested. The delimiter is not consumed (consume_delimiter=False)."""
        data = b'error: the command \\"python run.py\\" was not found" '
        value = b'error: the command \\"python run.py\\" was not found'
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b'"', b"\\")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        data = rb"^This is a simple regex string. It costs 10\$.$"
        value = rb"^This is a simple regex string. It costs 10\$."
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"$", b"\\")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        data = b"the searched file is .gitignore."
        value = b"the searched file is .gitignore"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b".", b" ")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

    def test6special_characters_escape_no_match(self):
        """In this test case special character escaping is tested without matching."""
        data = b'error: the command \\"python run.py\\" was not found\\" '
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b'"', b"\\")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = rb"^This is a simple regex string. It costs 10\$.\$"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"$", b"\\")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        data = b"the searched file is .gitignore ."
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b".", b" ")
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test7consume_delimiter(self):
        """In this test case check if the consume_delimiter parameter is working properly."""
        data = b"this is a match context.\n"

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"a", consume_delimiter=True)
        value = b"this is a"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"c", consume_delimiter=True)
        value = b"this is a matc"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"e", consume_delimiter=True)
        value = b"this is a match conte"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"\n", consume_delimiter=True)
        value = b"this is a match context.\n"
        match_context = DummyMatchContext(data)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"this is"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b" is", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"this"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"is", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"this is a match context.\n"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"context.\n", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

        value = b"this is a match context.\n"
        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"his is a match context.\n", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_match_results(data, match_element, match_context, self.id_, self.path, value, value, None)

    def test8consume_delimiter_no_match(self):
        """In this test case check if the consume_delimiter parameter is working properly and does not match data."""
        data = b"this is a match context.\n"

        for char in "bdfgjklpqruvwyz":
            delimited_data_model_element = DelimitedDataModelElement(self.id_, char.encode(), consume_delimiter=True)
            match_context = DummyMatchContext(data)
            match_element = delimited_data_model_element.get_match_element(self.path, match_context)
            self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"other data", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"isa", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"context\n", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

        match_context = DummyMatchContext(data)
        delimited_data_model_element = DelimitedDataModelElement(self.id_, b"this is a match context.\n", consume_delimiter=True)
        match_element = delimited_data_model_element.get_match_element(self.path, match_context)
        self.compare_no_match_results(data, match_element, match_context)

    def test9element_id_input_validation(self):
        """Check if element_id is validated."""
        self.assertRaises(ValueError, DelimitedDataModelElement, "", self.delimiter)  # empty element_id
        self.assertRaises(TypeError, DelimitedDataModelElement, None, self.delimiter)  # None element_id
        self.assertRaises(TypeError, DelimitedDataModelElement, b"path", self.delimiter)  # bytes element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, True, self.delimiter)  # boolean element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, 123, self.delimiter)  # integer element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, 123.22, self.delimiter)  # float element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, {"id": "path"}, self.delimiter)  # dict element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, ["path"], self.delimiter)  # list element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, [], self.delimiter)  # empty list element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, (), self.delimiter)  # empty tuple element_id is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, set(), self.delimiter)  # empty set element_id is not allowed

    def test10escape_input_validation(self):
        """Check if escape is validated."""
        self.assertRaises(ValueError, DelimitedDataModelElement, self.id_, self.delimiter, escape=b"")  # empty escape
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape="\\")  # string escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=True)  # boolean escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=123)  # integer escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=123.22)  # float escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape={"id": "path"})  # dict escape not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=["path"])  # list escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=[])  # empty list escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=())  # empty tuple escape is not allowed
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, escape=set())  # empty set escape is not allowed

    def test11consume_delimiter_input_validation(self):
        """Check if consume_delimiter is validated."""
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=b"")
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter="\\")
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=123)
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=123.22)
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter={"id": "path"})
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=["path"])
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=[])
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=())
        self.assertRaises(TypeError, DelimitedDataModelElement, self.id_, self.delimiter, consume_delimiter=set())

    def test12get_match_element_match_context_input_validation(self):
        """Check if an exception is raised, when other classes than MatchContext are used in get_match_element."""
        model_element = DelimitedDataModelElement(self.id_, self.delimiter)
        data = b"one, two, three"
        model_element.get_match_element(self.path, DummyMatchContext(data))
        model_element.get_match_element(self.path, MatchContext(data))

        self.assertRaises(AttributeError, model_element.get_match_element, self.path, MatchElement(None, data, None, None))
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, data.decode())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, True)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, 123.22)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, None)
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, [])
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, {"key": MatchContext(data)})
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, set())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, ())
        self.assertRaises(AttributeError, model_element.get_match_element, self.path, model_element)


if __name__ == "__main__":
    unittest.main()



"""
This module defines a model element that represents an IP address.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
import re
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ModelElementInterface import ModelElementInterface


class IpAddressDataModelElement(ModelElementInterface):
    """This class defines a model element that matches an IP address."""

    def __init__(self, element_id: str, ipv6: bool = False):
        """
        Create an element to match IP addresses.
        @param element_id an identifier for the ModelElement which is shown in the path.
        @param ipv6 if True, IPv6 addresses are parsed, IPv4 addresses are parsed otherwise.
        """
        super().__init__(element_id, ipv6=ipv6)

        if not ipv6:
            # self.regex = re.compile(br"((2[0-4][0-9]|1[0-9][0-9]|25[0-5]|[1-9]?[0-9])\.){3}(2[0-4][0-9]|1[0-9][0-9]|25[0-5]|[1-9]?[0-9])")
            # use a simpler regex to improve the performance.
            self.regex = re.compile(br"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}")
            self.extract = extract_ipv4_address
        else:
            # modified regex from https://community.helpsystems.com/forums/intermapper/miscellaneous-topics/
            # 5acc4fcf-fa83-e511-80cf-0050568460e4?_ga=2.113564423.1432958022.1523882681-2146416484.1523557976
            i4 = br"((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})"
            self.regex = re.compile(
                br"((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|"+i4+br"|:))|(([0-9A-Fa-f]{1,4}:"
                br"){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:"+i4+br"|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|((:[0-9A-Fa-f]{1,4})?"
                br":"+i4+br")|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,2}:"+i4+br")|:))|(([0-9A-Fa-f]{"
                br"1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:"+i4+br")|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{"
                br"1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:"+i4+br")|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,5}:"+i4+br")|:"
                br")))(%.+)?")
            self.extract = extract_ipv6_address

    def get_match_element(self, path: str, match_context):
        """
        Read an IP address at the current data position. When found, the match_object will be.
        Allowed formats for IPv6 addresses are defined in RFC4291 section 2.2.
        However, trailing IPv4 addresses (for example ::FFFF:129.144.52.38) are not allowed.
        """
        data = match_context.match_data
        m = self.regex.match(data)
        if m is None:
            return None
        match_len = m.span(0)[1]
        if self.extract is extract_ipv6_address and (b"." in m.group()[:match_len].split(b":")[-1] or (len(data) > match_len and (
                re.compile(br"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}").match(data[data.rfind(b":", 0, match_len) + 1:]) is not None or (
                data.find(b"::", match_len) == match_len and b"::" in data)))):
            return None
        extracted_address = self.extract(m.group(), match_len)
        if extracted_address is None:
            return None
        match_string = data[:match_len]
        match_context.update(match_string)
        return MatchElement(f"{path}/{self.element_id}", match_string, extracted_address, None)


def extract_ipv4_address(data: bytes, match_len: int):
    """Calculate integer values from ipv4 addresses."""
    numbers = [int(number) for number in data[:match_len].split(b".")]
    for number in numbers:
        if number > 255:
            return None
    return (numbers[0] << 24) + (numbers[1] << 16) + (numbers[2] << 8) + numbers[3]


def extract_ipv6_address(data: bytes, match_len: int):
    """Calculate integer values from ipv6 addresses."""
    parts = data[:match_len].split(b":")
    if b"" in parts:
        index = parts.index(b"")
        # addresses can start or end with ::. Handle this special case.
        parts = [number for number in parts if number != b""]
        parts = parts[:index] + [b"0"] * (8 - len(parts)) + parts[index:]
    numbers = [int(b"0x" + number, 16) for number in parts]
    for number in numbers:
        if number > 65535:
            return None
    return (numbers[0] << 112) + (numbers[1] << 96) + (numbers[2] << 80) + (numbers[3] << 64) + (numbers[4] << 48) + (numbers[5] << 32)\
        + (numbers[6] << 16) + (numbers[7])



from time import sleep
import socket
sock_name = '/tmp/test5unixSocket.sock'  # skipcq: BAN-B108

sleep(0.5)
client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
client.connect(sock_name)
client.send(b'data')
client.close()



"""This module provides support for parsing of sshd messages."""

from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement
from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement


def get_model(user_name_model=None):
    """Return a model to parse a sshd information message after any standard logging preamble, e.g. from syslog."""
    if user_name_model is None:
        user_name_model = VariableByteDataModelElement("user", b"0123456789abcdefghijklmnopqrstuvwxyz.-")

    from_str = b" from "
    port = b" port "
    preauth = b" [preauth]"

    type_children = [
        SequenceModelElement("accepted key", [
            FixedDataModelElement("s0", b"Accepted publickey for "),
            user_name_model,
            FixedDataModelElement("s1", from_str),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s2", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s3", b" ssh2: "),
            DelimitedDataModelElement("asym-algorithm", b" ", consume_delimiter=True),
            VariableByteDataModelElement("fingerprint", b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+/:"),
            OptionalMatchModelElement("opt", SequenceModelElement("seq", [
                FixedDataModelElement("s4", b" ID "),
                DelimitedDataModelElement("id", b" "),
                FixedDataModelElement("s5", b" (serial "),
                DecimalIntegerValueModelElement("serial"),
                FixedDataModelElement("s6", b") CA "),
                AnyByteDataModelElement("algorithm_details")
            ]))
        ]),
        SequenceModelElement("btmp-perm", [
            FixedDataModelElement("s0", b"Excess permission or bad ownership on file /var/log/btmp")
        ]),
        SequenceModelElement("close-sess", [
            FixedDataModelElement("s0", b"Close session: user "),
            user_name_model,
            FixedDataModelElement("s1", from_str),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s2", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s3", b" id "),
            DecimalIntegerValueModelElement("userid")
        ]),
        SequenceModelElement("closing", [
            FixedDataModelElement("s0", b"Closing connection to "),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s1", port),
            DecimalIntegerValueModelElement("port")
        ]),
        SequenceModelElement("closed", [
            FixedDataModelElement("s0", b"Connection closed by "),
            FirstMatchModelElement("fm", [
                IpAddressDataModelElement("clientip"),
                SequenceModelElement("seq", [
                    FixedWordlistDataModelElement("user-type", [b"authenticating", b"invalid"]),
                    FixedDataModelElement("s1", b" user "),
                    DelimitedDataModelElement("user", b" "),
                    FixedDataModelElement("s2", b" "),
                    FirstMatchModelElement("fm", [
                        IpAddressDataModelElement("ip"),
                        IpAddressDataModelElement("ipv6", ipv6=True)
                    ]),
                    FixedDataModelElement("s3", b" port "),
                    DecimalIntegerValueModelElement("port"),
                    FixedDataModelElement("s4", b" [preauth]")
                ])
            ])
        ]),
        SequenceModelElement("connect", [
            FixedDataModelElement("s0", b"Connection from "),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s1", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s2", b" on "),
            IpAddressDataModelElement("serverip"),
            FixedDataModelElement("s3", port),
            DecimalIntegerValueModelElement("sport")
        ]),
        SequenceModelElement("disconnectreq", [
            FixedDataModelElement("s0", b"Received disconnect from "),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s1", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s2", b":"),
            DecimalIntegerValueModelElement("session"),
            FixedDataModelElement("s3", b": "),
            FixedWordlistDataModelElement("reason", [b"disconnected by user"])
        ]),
        SequenceModelElement("disconnected", [
            FixedDataModelElement("s0", b"Disconnected from "),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s1", port),
            DecimalIntegerValueModelElement("port")
        ]),
        FixedDataModelElement("error-bind",  b"error: bind: Cannot assign requested address"),
        SequenceModelElement("error-max-auth", [
            FixedDataModelElement("s0", b"error: maximum authentication attempts exceeded for "),
            OptionalMatchModelElement("opt", FixedDataModelElement("invalid", b"invalid user ")),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s1", b" from "),
            FirstMatchModelElement("fm", [
                IpAddressDataModelElement("from_ip"),
                IpAddressDataModelElement("from_ip_v6", ipv6=True)
            ]),
            FixedDataModelElement("s2", b" port "),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s3", b" ssh2 [preauth]")
        ]),
        FixedDataModelElement("error-cert-exp", b"error: Certificate invalid: expired"),
        FixedDataModelElement("error-cert-not-yet-valid", b"error: Certificate invalid: not yet valid"),
        FixedDataModelElement("error-cert-not-listed-principal", b"error: Certificate invalid: name is not a listed principal"),
        FixedDataModelElement("error-refused-by-options", b"error: Refused by certificate options"),
        SequenceModelElement("error-channel-setup", [
            FixedDataModelElement("s0", b"error: channel_setup_fwd_listener: cannot listen to port: "),
            DecimalIntegerValueModelElement("port")
        ]),
        SequenceModelElement("error-auth-key", [
            FixedDataModelElement("s0", b"error: Authentication key "),
            DelimitedDataModelElement("asym-algorithm", b" "),
            FixedDataModelElement("s1", b" "),
            DelimitedDataModelElement("hash", b" "),
            FixedDataModelElement("s2", b" revoked by file "),
            AnyByteDataModelElement("file")
        ]),
        SequenceModelElement("error-load-key", [
            FixedDataModelElement("s0", b"error: Could not load host key: "),
            AnyByteDataModelElement("file")
        ]),
        SequenceModelElement("ident-missing", [
            FixedDataModelElement("s0", b"Did not receive identification string from "),
            IpAddressDataModelElement("clientip")
        ]),
        SequenceModelElement("invalid-user", [
            FixedDataModelElement("s0", b"Invalid user "),
            DelimitedDataModelElement("user", from_str),
            FixedDataModelElement("s1", from_str),
            FirstMatchModelElement("fm", [
                IpAddressDataModelElement("from_ip"),
                IpAddressDataModelElement("from_ip_v6", ipv6=True)
            ]),
            FixedDataModelElement("s2", b" port "),
            DecimalIntegerValueModelElement("port")
        ]),
        SequenceModelElement("invalid-user-auth-req", [
            FixedDataModelElement("s0", b"input_userauth_request: invalid user "),
            DelimitedDataModelElement("user", preauth),
            FixedDataModelElement("s1", preauth)
        ]),
        SequenceModelElement("postppk", [
            FixedDataModelElement("s0", b"Postponed publickey for "),
            user_name_model,
            FixedDataModelElement("s1", from_str),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s2", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s3", b" ssh2 [preauth]")
        ]),
        SequenceModelElement("readerr", [
            FixedDataModelElement("s0", b"Read error from remote host "),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s1", b": Connection timed out")
        ]),
        SequenceModelElement("disconnect", [
            FixedDataModelElement("s0", b"Received disconnect from "),
            FirstMatchModelElement("fm", [
                IpAddressDataModelElement("from_ip"),
                IpAddressDataModelElement("from_ip_v6", ipv6=True)
            ]),
            FixedDataModelElement("s1", b": 11: "),
            FirstMatchModelElement("reason", [
                FixedDataModelElement("disconnected", b"disconnected by user"),
                SequenceModelElement("remotemsg", [
                    DelimitedDataModelElement("msg", preauth),
                    FixedDataModelElement("s0", preauth)
                ])
            ])
        ]),
        SequenceModelElement("signal", [
            FixedDataModelElement("s0", b"Received signal "),
            DecimalIntegerValueModelElement("signal"),
            FixedDataModelElement("s1", b"; terminating.")
        ]),
        SequenceModelElement("server", [
            FixedDataModelElement("s0", b"Server listening on "),
            DelimitedDataModelElement("serverip", b" "),
            FixedDataModelElement("s1", port),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s2", b".")
        ]),
        SequenceModelElement("oom-adjust", [
            FixedDataModelElement("s0", b"Set /proc/self/oom_score_adj "),
            OptionalMatchModelElement("from", FixedDataModelElement("default", b"from 0 ")),
            FixedDataModelElement("s1", b"to "),
            DecimalIntegerValueModelElement("newval", value_sign_type=DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL)
        ]),
        SequenceModelElement("session-start", [
            FixedDataModelElement("s0", b"Starting session: "),
            FirstMatchModelElement("sess-info", [
                SequenceModelElement("shell", [
                    FixedDataModelElement("s0", b"shell on "),
                    DelimitedDataModelElement("terminal", b" ")
                ]),
                SequenceModelElement("subsystem", [
                    FixedDataModelElement("s0", b"subsystem \"sftp\"")
                ]),
                SequenceModelElement("forced-command", [
                    FixedDataModelElement("s0", b"forced-command (key-option) \""),
                    DelimitedDataModelElement("command", b"\" for "),
                    FixedDataModelElement("s1", b"\"")
                ])
            ]),
            FixedDataModelElement("s1", b" for "),
            user_name_model,
            FixedDataModelElement("s2", from_str),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s3", port),
            DecimalIntegerValueModelElement("port"),
            OptionalMatchModelElement("idinfo", SequenceModelElement("idinfo", [
                FixedDataModelElement("s0", b" id "),
                DecimalIntegerValueModelElement("id")
            ]))
        ]),
        SequenceModelElement("transferred", [
            FixedDataModelElement("s0", b"Transferred: sent "),
            DecimalIntegerValueModelElement("sent"),
            FixedDataModelElement("s1", b", received "),
            DecimalIntegerValueModelElement("received"),
            FixedDataModelElement("s1", b" bytes")]),
        SequenceModelElement("pam", [
            FixedDataModelElement("s0", b"pam_unix(sshd:session): session "),
            FixedWordlistDataModelElement("change", [b"opened", b"closed"]),
            FixedDataModelElement("s1", b" for user "),
            user_name_model,
            OptionalMatchModelElement("openby", FixedDataModelElement("default", b" by (uid=0)"))
        ]),
        SequenceModelElement("child", [
            FixedDataModelElement("s0", b"User child is on pid "),
            DecimalIntegerValueModelElement("pid")
        ]),
        SequenceModelElement("failed/accept", [
            FixedWordlistDataModelElement("s0", [b"Failed ", b"Accepted "]),
            FixedWordlistDataModelElement("type", [b"password", b"none", b"publickey"]),
            FixedDataModelElement("s1", b" for "),
            OptionalMatchModelElement("opt", FixedDataModelElement("invalid", b"invalid user ")),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s2", b" from "),
            FirstMatchModelElement("fm", [
                IpAddressDataModelElement("from_ip"),
                IpAddressDataModelElement("from_ip_v6", ipv6=True)
            ]),
            FixedDataModelElement("s3", b" port "),
            DecimalIntegerValueModelElement("port"),
            AnyByteDataModelElement("service")
        ]),
        SequenceModelElement("disconnecting", [
            FixedDataModelElement("s0", b"Disconnecting "),
            FixedWordlistDataModelElement("type", [b"authenticating", b"invalid"]),
            FixedDataModelElement("s1", b" user "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s1", b" "),
            IpAddressDataModelElement("ip"),
            FixedDataModelElement("s2", b" port "),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s3", b": Too many authentication failures [preauth]")
        ]),
        SequenceModelElement("fatal", [
            FixedDataModelElement("s0", b"fatal: Timeout before authentication for "),
            IpAddressDataModelElement("ip"),
            FixedDataModelElement("s1", b" port "),
            DecimalIntegerValueModelElement("port")
        ]),
        SequenceModelElement("cert-auth", [
            FixedDataModelElement("s0", b"cert: Authentication tried for "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s1", b" with valid certificate but not from a permitted source address ("),
            IpAddressDataModelElement("ip"),
            FixedDataModelElement("s2", b")."),
        ]),
        SequenceModelElement("change-root-dir", [
            FixedDataModelElement("s0", b"Changed root directory to \""),
            DelimitedDataModelElement("root-dir", b"\""),
            FixedDataModelElement("s1", b"\"")
        ]),
        FixedDataModelElement("subsystem-request", b"subsystem request for sftp"),
        SequenceModelElement("conn-write-poll", [
            FixedDataModelElement("s0", b"packet_write_poll: Connection from "),
            IpAddressDataModelElement("from_ip"),
            FixedDataModelElement("s1", b" port "),
            DecimalIntegerValueModelElement("port"),
            FixedDataModelElement("s2", b": Host is down")
        ]),
        SequenceModelElement("debug", [
            FixedDataModelElement("s0", b"debug"),
            DecimalIntegerValueModelElement("debug-num"),
            FixedDataModelElement("s1", b": "),
            FirstMatchModelElement("fm", [
                SequenceModelElement("seq1", [
                    FixedDataModelElement("s2", b"Got "),
                    DecimalIntegerValueModelElement("num1"),
                    FixedDataModelElement("s3", b"/"),
                    DecimalIntegerValueModelElement("num2"),
                    FixedDataModelElement("s4", b" for keepalive")
                ]),
                SequenceModelElement("seq2", [
                    FixedDataModelElement("s2", b"channel "),
                    DecimalIntegerValueModelElement("channel-num"),
                    FixedDataModelElement("s3", b": request "),
                    DelimitedDataModelElement("mail", b" "),
                    FixedDataModelElement("s4", b" confirm "),
                    DecimalIntegerValueModelElement("num")
                ]),
                SequenceModelElement("seq3", [
                    FixedDataModelElement("s2", b"send packet: type "),
                    DecimalIntegerValueModelElement("packet-type")
                ]),
                SequenceModelElement("seq4", [
                    FixedDataModelElement("s2", b"receive packet: type "),
                    DecimalIntegerValueModelElement("packet-type")
                ]),
                FixedDataModelElement("do-cleanup", b"do_cleanup"),
                SequenceModelElement("seq5", [
                    FixedDataModelElement("s2", b"session_pty_cleanup: session "),
                    DecimalIntegerValueModelElement("sess-num"),
                    FixedDataModelElement("s3", b" release "),
                    AnyByteDataModelElement("file")
                ])
            ])
        ]),
        SequenceModelElement("pam_succeed_if", [
            FixedDataModelElement("s0", b"pam_succeed_if(sshd:auth): requirement \"uid >= "),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s1", b"\" not met by user \""),
            DelimitedDataModelElement("user", b"\""),
            FixedDataModelElement("s2", b"\"")
        ]),
    ]

    model = SequenceModelElement("sshd", [
        FixedDataModelElement("sname", b"sshd["),
        DecimalIntegerValueModelElement("pid"),
        FixedDataModelElement("s0", b"]: "),
        FirstMatchModelElement("msg", type_children)
    ])
    return model



"""
This module defines a detector for log atoms not matching any allowlisted rule.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer.AminerConfig import CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig


class AllowlistViolationDetector(AtomHandlerInterface):
    """
    Objects of this class handle a list of allowlist rules.
    They ensure, that each received log-atom is at least covered by a single allowlist rule. To avoid traversing the complete rule tree
    more than once, the allowlist rules may have match actions attached that set off an alarm by themselves.
    """

    def __init__(self, aminer_config, allowlist_rules, anomaly_event_handlers, output_logline=True):
        """
        Initialize the detector.
        @param allowlist_rules list of rules executed until the first rule matches.
        """
        super().__init__(aminer_config=aminer_config, anomaly_event_handlers=anomaly_event_handlers, output_logline=output_logline,
                         allowlist_rules=allowlist_rules)

    def receive_atom(self, log_atom):
        """
        Receive a parsed atom and the information about the parser match.
        @param log_atom atom with parsed data to check
        @return a boolean value if the log atom matches one of the rules.
        """
        self.log_total += 1
        event_data = {}
        for rule in self.allowlist_rules:
            if rule.match(log_atom):
                self.log_success += 1
                return True
        original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
        try:
            data = log_atom.raw_data.decode(AminerConfig.ENCODING)
        except UnicodeError:
            data = repr(log_atom.raw_data)
        analysis_component = {'AffectedLogAtomPaths': list(log_atom.parser_match.get_match_dictionary()), 'AffectedLogAtomValues': [data]}
        sorted_log_lines = [original_log_line_prefix + data]
        event_data['AnalysisComponent'] = analysis_component
        for listener in self.anomaly_event_handlers:
            listener.receive_event(f'Analysis.{self.__class__.__name__}', 'No allowlisting for current atom', sorted_log_lines,
                                   event_data, log_atom, self)
        return False

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        super().log_statistics(component_name)
        for i, rule in enumerate(self.allowlist_rules):
            rule.log_statistics(component_name + '.' + rule.__class__.__name__ + str(i))



import unittest
from aminer.analysis.EventSequenceDetector import EventSequenceDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ParserMatch import ParserMatch
from unit.TestBase import TestBase


class TestHandler():
    """Dummy anomaly handler."""

    def __init__(self):
        self.anomaly = None

    # skipcq: PYL-W0613
    def receive_event(self, name, msg, ll, evdat, atom, obj):
        """Receive anomaly information."""
        self.anomaly = evdat


class EventSequenceDetectorTest(TestBase):
    """Unittests for the EventSequenceDetectorDetector."""

    def test1_normal_sequence_detection(self):
        """
        This test case checks the normal detection of new sequences.
        The ESD is used to detect value sequences of length 2 and uses one id path to cope with
        interleaving sequences, i.e., the sequences only make sense when logs that contain the
        same id are considered.
        """
        description = "Test1EventSequenceDetector"

        # Initialize detector for sequence length 2
        test_handler = TestHandler()
        event_sequence_detector = EventSequenceDetector(self.aminer_config, [test_handler], ['/model/id'], ['/model/value'], 2, False, -1,
                                                        'Default', True, output_logline=False)
        self.analysis_context.register_component(event_sequence_detector, description)

        # Prepare log atoms that represent two users (id) that produce interleaved sequence a, b, c
        # This means, user with id 1 creates sequence a, b, c, and user with id 2 creates sequence
        # a, b, however, these sequences are interleaved. The ESD resolves this issue using the id
        # as an id path (/model/id). The path of the values is /model/value.
        # The following events are generated:
        #  id: 1 value: a
        #  id: 1 value: b
        #  id: 2 value: a
        #  id: 1 value: c
        #  id: 2 value: b
        m_1 = MatchElement('/model/id', b'1', b'1', None)
        m_2 = MatchElement('/model/value', b'a', b'a', None)
        match_element_1 = MatchElement('/model', b'1a', b'1a', [m_1, m_2])
        parser_match_1 = ParserMatch(match_element_1)
        log_atom_1 = LogAtom(b'1a', parser_match_1, 1, None)

        m_3 = MatchElement('/model/id', b'1', b'1', None)
        m_4 = MatchElement('/model/value', b'b', b'b', None)
        match_element_2 = MatchElement('/model', b'1b', b'1b', [m_3, m_4])
        parser_match_2 = ParserMatch(match_element_2)
        log_atom_2 = LogAtom(b'1b', parser_match_2, 2, None)

        m_5 = MatchElement('/model/id', b'2', b'2', None)
        m_6 = MatchElement('/model/value', b'a', b'a', None)
        match_element_3 = MatchElement('/model', b'2a', b'2a', [m_5, m_6])
        parser_match_3 = ParserMatch(match_element_3)
        log_atom_3 = LogAtom(b'2a', parser_match_3, 3, None)

        m_7 = MatchElement('/model/id', b'1', b'1', None)
        m_8 = MatchElement('/model/value', b'c', b'c', None)
        match_element_4 = MatchElement('/model', b'1c', b'1c', [m_7, m_8])
        parser_match_4 = ParserMatch(match_element_4)
        log_atom_4 = LogAtom(b'1c', parser_match_4, 4, None)

        m_9 = MatchElement('/model/id', b'2', b'2', None)
        m_10 = MatchElement('/model/value', b'b', b'b', None)
        match_element_5 = MatchElement('/model', b'2b', b'2b', [m_9, m_10])
        parser_match_5 = ParserMatch(match_element_5)
        log_atom_5 = LogAtom(b'2b', parser_match_5, 5, None)

        # Forward log atoms to detector
        # Since sequence length is 2, first atom should not have any effect
        # Input: id: 1 value: a
        # Expected output: None
        event_sequence_detector.receive_atom(log_atom_1)
        self.assertIsNone(test_handler.anomaly)
        sequences_set = set()
        self.assertEqual(event_sequence_detector.sequences, sequences_set)

        # Second log atom should create first sequence
        # Input: id: 1 value: b
        # Expected output: New sequence (a, b) detected, added to known sequences
        event_sequence_detector.receive_atom(log_atom_2)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/model/value'],
                                                                      'AffectedLogAtomValues': [('a',), ('b',)],
                                                                      'AffectedIdValues': ['1']}})
        sequences_set.add((('a',), ('b',)))
        self.assertEqual(event_sequence_detector.sequences, sequences_set)
        test_handler.anomaly = None

        # Next log atom is of different user, should not have any effect
        # Input: id: 2 value: a
        # Expected output: None
        event_sequence_detector.receive_atom(log_atom_3)
        self.assertIsNone(test_handler.anomaly)
        self.assertEqual(event_sequence_detector.sequences, sequences_set)

        # Next log atom is of user with id 1, but new value c, thus new sequence should be generated
        # Input: id: 1 value: c
        # Expected output: New sequence (b, c) detected, added to known sequences
        event_sequence_detector.receive_atom(log_atom_4)
        self.assertEqual(test_handler.anomaly, {'AnalysisComponent': {'AffectedLogAtomPaths': ['/model/value'],
                                                                      'AffectedLogAtomValues': [('b',), ('c',)],
                                                                      'AffectedIdValues': ['1']}})
        sequences_set.add((('b',), ('c',)))
        self.assertEqual(event_sequence_detector.sequences, sequences_set)
        test_handler.anomaly = None

        # Next log atom is of user with id 2, but sequence a, b is already known from user with id 1, thus no effect
        # Input: id: 2 value: b
        # Expected output: None
        event_sequence_detector.receive_atom(log_atom_5)
        self.assertIsNone(test_handler.anomaly)
        self.assertEqual(event_sequence_detector.sequences, sequences_set)


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
from time import time
from aminer.analysis.TimestampsUnsortedDetector import TimestampsUnsortedDetector
from unit.TestBase import TestBase
from datetime import datetime


class TimestampsUnsortedDetectorTest(TestBase):
    """Unittests for the TimestampsUnsortedDetector."""

    __expected_string = '%s Timestamp %s below %s\n%s: "%s" (%d lines)\n  %s\n\n'
    pid = b' pid='
    datetime_format_string = '%Y-%m-%d %H:%M:%S'

    def test1timestamp_lower_than_last_timestamp(self):
        """This test case checks if an event is created, when the timestamp is lower than the last one."""
        description = "Test1TimestampsUnsortedDetector"
        match_context_fixed_dme = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element_fixed_dme = fixed_dme.get_match_element("match", match_context_fixed_dme)
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)
        self.analysis_context.register_component(new_match_path_detector, description)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element_fixed_dme), t, new_match_path_detector)
        timestamp_unsorted_detector = TimestampsUnsortedDetector(self.aminer_config, [self.stream_printer_event_handler], False,
                                                                 output_logline=False)
        self.analysis_context.register_component(timestamp_unsorted_detector, description + "2")
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), '')

        log_atom.set_timestamp(t - 10000)
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t - 10000).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t - 10000).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), timestamp_unsorted_detector.__class__.__name__,
            description + "2", 1, " pid="))

    def test2timestamp_lower_than_last_timestamp_exit_on_error(self):
        """This test case checks if the program exits, when the timestamp is lower than the last one and the exitOnError flag is set."""
        description = "Test2TimestampsUnsortedDetector"
        match_context_fixed_dme = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element_fixed_dme = fixed_dme.get_match_element("match", match_context_fixed_dme)
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)
        self.analysis_context.register_component(new_match_path_detector, description)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element_fixed_dme), t, new_match_path_detector)
        timestamp_unsorted_detector = TimestampsUnsortedDetector(self.aminer_config, [self.stream_printer_event_handler], True,
                                                                 output_logline=False)
        self.analysis_context.register_component(timestamp_unsorted_detector, description + "2")
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), '')

        log_atom.set_timestamp(t - 10000)
        with self.assertRaises(SystemExit) as cm:
            timestamp_unsorted_detector.receive_atom(log_atom)
        self.assertEqual(cm.exception.code, 1)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t - 10000).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t - 10000).strftime(self.datetime_format_string),
            datetime.fromtimestamp(t).strftime(self.datetime_format_string), timestamp_unsorted_detector.__class__.__name__,
            description + "2", 1, " pid="))

    def test3timestamp_higher_than_last_timestamp(self):
        """This test case checks if nothing happens, when the timestamp is, as expected, higher than the last one."""
        description = "Test3TimestampsUnsortedDetector"
        match_context_fixed_dme = MatchContext(self.pid)
        fixed_dme = FixedDataModelElement('s1', self.pid)
        match_element_fixed_dme = fixed_dme.get_match_element("match", match_context_fixed_dme)
        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False)
        self.analysis_context.register_component(new_match_path_detector, description)

        t = time()
        log_atom = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element_fixed_dme), t, new_match_path_detector)
        timestamp_unsorted_detector = TimestampsUnsortedDetector(self.aminer_config, [self.stream_printer_event_handler], False,
                                                                 output_logline=False)
        self.analysis_context.register_component(timestamp_unsorted_detector, description + "2")
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), '')

        log_atom.set_timestamp(t)
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), '')

        log_atom.set_timestamp(t + 10000)
        self.assertTrue(timestamp_unsorted_detector.receive_atom(log_atom))
        self.assertEqual(self.output_stream.getvalue(), '')


if __name__ == "__main__":
    unittest.main()



"""This module defines a parser for cron."""

from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement


def get_model(user_name_model=None):
    """Return a model to parse a cron message logged via syslog after any standard logging preamble, e.g. from syslog."""
    if user_name_model is None:
        user_name_model = VariableByteDataModelElement("user", b"0123456789abcdefghijklmnopqrstuvwxyz.-")

    type_children = [
        SequenceModelElement("exec", [
            FixedDataModelElement("s0", b"("),
            user_name_model,
            FixedDataModelElement("s1", b") CMD "),
            AnyByteDataModelElement("command")
        ]),
        SequenceModelElement("pam", [
            FixedDataModelElement("s0", b"pam_unix(cron:session): session "),
            FixedWordlistDataModelElement("change", [b"opened", b"closed"]),
            FixedDataModelElement("s1", b" for user "),
            user_name_model,
            OptionalMatchModelElement("openby", FixedDataModelElement("default", b" by (uid=0)"))
        ])
    ]

    model = FirstMatchModelElement("cron", [
        SequenceModelElement("std", [
            FixedDataModelElement("sname", b"CRON["),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s0", b"]: "),
            FirstMatchModelElement("msgtype", type_children)
        ]),
        SequenceModelElement("low", [
            FixedDataModelElement("sname", b"cron["),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s0", b"]: (*system*"),
            DelimitedDataModelElement("rname", b") RELOAD ("),
            FixedDataModelElement("s1", b") RELOAD ("),
            DelimitedDataModelElement("fname", b")"),
            FixedDataModelElement("s2", b")"), ])
    ])
    return model
